# -*- coding: utf-8 -*-
from icemac.addressbook.file.interfaces import IFile
from icemac.addressbook.interfaces import IEntityOrder, IEntity
from icemac.addressbook.interfaces import IPerson, IPostalAddress
from icemac.addressbook.testing import set_modified, delete_field
from zope.dublincore.interfaces import IZopeDublinCore
from zope.testbrowser.browser import HTTPError
import gocept.country.db
import pytest
import zope.component.hooks
import zope.testbrowser.interfaces


def form_fields_names(browser):
    """Get a list of field names as "<entity name>.<field name>"."""
    fields = ['.'.join(x.replace('IcemacAddressbookPerson', '')
                       .replace('IcemacAddressbookAddress', '')
                       for x in x.name.split('.')[-2:])
              for x in browser.getForm().controls
              if not zope.testbrowser.interfaces.IItemControl.providedBy(x)]
    # Omit empty marker hidden fields as they are too technical:
    return [x for x in fields if not x.endswith('-empty-marker')]


def test_person__PersonAddForm__1(address_book, browser):
    """`PersonAddForm` allows editors to add new persons."""
    browser.login('editor')
    browser.open(browser.PERSONS_LIST_URL)
    browser.getLink('person').click()
    assert browser.PERSON_ADD_URL == browser.url
    # There are fields for the person's personal data:
    browser.getControl('first name').value = u'Hans'
    browser.getControl('last name').value = u'Tester'
    browser.getControl('birth date').value = u'2000 2 29 '
    # Fields for the primary postal address:
    browser.getControl('address prefix').value = u'c/o Mama'
    browser.getControl('street').value = u'Demoweg 23'
    browser.getControl('city', index=0).value = u'Testhausen'
    browser.getControl('zip').value = u'99999'
    browser.getControl('country').displayValue = ['Austria']
    # The primary phone number:
    browser.getControl('number').value = u'112'
    # The primary e-mail address:
    browser.getControl('e-mail address').value = 'hans@example.com'
    # The primary home page address:
    browser.getControl('URL').value = 'http://www.example.com'
    # After filling out the fields correctly, we can add the record to the
    # database and get back to the address book where the entry is
    # displayed:
    browser.getControl('Add').click()
    assert '"Tester, Hans" added.' == browser.message
    assert browser.PERSONS_LIST_URL == browser.url
    assert (['Tester', 'Hans'] ==
            browser.etree.xpath('//table/tbody/tr/td/a/text()'))


def test_person__PersonAddForm__2(address_book, browser):
    """`PersonAddForm` cannot be accessed by a visitor."""
    browser.login('visitor')
    with pytest.raises(HTTPError) as err:
        browser.open(browser.PERSON_ADD_URL)
    assert 'HTTP Error 403: Forbidden' == str(err.value)


def test_person__PersonAddForm__3(address_book, browser):
    """`PersonAddForm` validates the input fields."""
    browser.login('editor')
    browser.open(browser.PERSON_ADD_URL)
    # Person and address data are relatively complex. There are some
    # required fields which lead to an error message when they are not
    # filled. Some fields get parsed, so entering rubbish leads also to an
    # error message. (Note: Germany is by default selected as country.
    # Clicking on it, deselects it, so no country is selected afterwards.):
    browser.getControl('birth date').value = u'qqqqq'
    browser.getControl('e-mail address').value = u'qqqqq'
    browser.getControl('URL').value = 'qqqqq'
    browser.getControl('Germany').click()  # deselects the default selection
    browser.getControl('Add').click()
    assert [] == browser.message
    assert browser.PERSON_ADD_URL == browser.url
    assert (['Required input is missing.',
             "The datetime string did not match the pattern u'yyyy MMM d '.",
             'Constraint not satisfied',
             'The specified URI is not valid.'] ==
            browser.etree.xpath('//ul[@class="errors"]/li/div/text()'))


def test_person__PersonAddForm__4(address_book, FieldFactory, browser):
    """User defined fields show up in the person add form."""
    FieldFactory(address_book, IPerson, 'TextLine', u'baz')
    browser.login('editor')
    browser.open(browser.PERSON_ADD_URL)
    browser.getControl('last name').value = 'Tester'
    assert '' == browser.getControl('baz').value
    browser.getControl('baz').value = 'monster'
    browser.getControl('Add').click()
    assert '"Tester" added.' == browser.message
    assert browser.PERSONS_LIST_URL == browser.url


def test_person__PersonAddForm__5(address_book, FieldFactory, browser):
    """An `editor` can enter values in user defined field in the add form."""
    FieldFactory(address_book, IPerson, 'TextLine', u'foobar')
    browser.login('editor')
    browser.open(browser.PERSON_ADD_URL)
    browser.getControl('last name').value = 'Bester'
    assert '' == browser.getControl('foobar').value
    browser.getControl('foobar').value = 'better'
    browser.getControl('Add').click()
    assert '"Bester" added.' == browser.message
    assert browser.PERSONS_LIST_URL == browser.url
    # The entered value got persisted:
    browser.getLink('Bester').click()
    assert browser.PERSON_EDIT_URL == browser.url
    assert 'better' == browser.getControl('foobar').value


@pytest.mark.parametrize('username', ('editor', 'mgr'))
def test_person__PersonAddForm__6(
        address_book, FieldFactory, browser2, browser, username):
    """A deleted field is not displayed any more in the add form."""
    delete_field(
        browser2,
        FieldFactory(address_book, IPerson, 'TextLine', u'foobar').__name__)
    browser.login(username)
    browser.open(browser.PERSON_ADD_URL)
    with pytest.raises(LookupError):
        browser.getControl('foobar')


def test_person__PersonAddForm__7(address_book, KeywordFactory, browser):
    """A keyword can be assigned using `PersonAddForm`."""
    KeywordFactory(address_book, u'friend')
    KeywordFactory(address_book, u'Company co-worker')
    family = KeywordFactory(address_book, u'family')
    church = KeywordFactory(address_book, u'church')
    browser.login('editor')
    browser.open(browser.PERSON_ADD_URL)
    browser.getControl('last name').value = 'keyword test'
    # In browser tests keywords are selected using a simple multiselect widget
    # because the testbrowser does not support JavaScript. The widget currently
    # displays all existing keywords case insensitive alphabetically sorted.
    assert ['church',
            'Company co-worker',
            'family',
            'friend'] == browser.getControl('keywords').displayOptions
    # Keyword assignments are stored when saving:
    browser.getControl('church').click()
    browser.getControl('family').click()
    browser.getControl('Add').click()
    assert '"keyword test" added.' == browser.message
    with zope.component.hooks.site(address_book):
        assert set([church, family]) == set(address_book['Person'].keywords)


def test_person__PersonAddForm__8(address_book, browser):
    """`PersonAddForm` uses the default entity and field sort order."""
    browser.login('editor')
    browser.open(browser.PERSON_ADD_URL)
    assert ['Person.first_name',
            'Person.last_name',
            'Person.birth_date',
            'Person.keywords:list',
            'Person.notes',
            'Postaladdress.address_prefix',
            'Postaladdress.street',
            'Postaladdress.city',
            'Postaladdress.zip',
            'Postaladdress.country:list',
            'Phonenumber.number',
            'Emailaddress.email',
            'Homepageaddress.url',
            'buttons.add',
            'buttons.cancel'] == form_fields_names(browser)


def test_person__PersonAddForm__9(address_book, browser):
    """`PersonAddForm` respects the the user defined entity sort order."""
    browser.login('editor')
    postal_address_entity = IEntity(IPostalAddress)
    order = zope.component.getUtility(IEntityOrder)
    # When moving the postal address two positions up, the sort order in the
    # person add form changes accordingly:
    order.up(postal_address_entity, delta=2)
    browser.open(browser.PERSON_ADD_URL)
    assert ['Postaladdress.address_prefix',
            'Postaladdress.street',
            'Postaladdress.city',
            'Postaladdress.zip',
            'Postaladdress.country:list',
            'Person.first_name',
            'Person.last_name',
            'Person.birth_date',
            'Person.keywords:list',
            'Person.notes',
            'Phonenumber.number',
            'Emailaddress.email',
            'Homepageaddress.url',
            'buttons.add',
            'buttons.cancel'] == form_fields_names(browser)


def test_person__PersonAddForm__10(translated_address_book, browser):
    """`PersonAddForm` translates country names."""
    browser.login('editor')
    browser.lang('de-DE')
    browser.open(browser.PERSON_ADD_URL)
    assert ['Deutschland'] == browser.getControl('Land').displayValue
    browser.getControl('Land').displayValue = ['Schweiz']
    browser.getControl('Ort').value = 'Thun'
    browser.getControl('Familienname').value = 'Deutsch'
    browser.getControl(u'Hinzufügen').click()
    assert u'„Deutsch“ hinzugefügt.' == browser.message


def test_person__PersonAddForm__11(translated_address_book, browser):
    """`PersonAddForm` translates error messages using `zope.app.locales`."""
    # This test makes sure that `zope.app.locales` is installed.
    browser.login('editor')
    browser.lang('de-DE')
    browser.open(browser.PERSON_ADD_URL)
    browser.getControl(u'Hinzufügen').click()
    assert ['Erforderliche Eingabe fehlt.'] == browser.etree.xpath(
        '//ul[@class="errors"]/li/div/text()')


def test_person__AddGroup__1(address_book, browser):
    """`AddGroup` respects the user definded field sort order."""
    browser.login('editor')
    person_entity = IEntity(IPerson)
    # When switching the first name and the last name of the person the person
    # add form changes accordingly:
    person_entity.setFieldOrder(('last_name', 'first_name'))
    browser.open(browser.PERSON_ADD_URL)
    assert ['Person.last_name',
            'Person.first_name',
            'Person.birth_date',
            'Person.keywords:list',
            'Person.notes',
            'Postaladdress.address_prefix',
            'Postaladdress.street',
            'Postaladdress.city',
            'Postaladdress.zip',
            'Postaladdress.country:list',
            'Phonenumber.number',
            'Emailaddress.email',
            'Homepageaddress.url',
            'buttons.add',
            'buttons.cancel'] == form_fields_names(browser)


def test_person__PersonEditForm__1(person_data, browser):
    """`PersonEditForm` allows to change person's data."""
    browser.login('editor')
    browser.open(browser.PERSONS_LIST_URL)
    browser.getLink('Tester').click()
    # The table entry is a link which leads to the edit page of the person
    # where the data of the person is displayed:
    assert browser.PERSON_EDIT_URL == browser.url
    assert 'Petra' == browser.getControl('first name').value
    assert 'Tester' == browser.getControl('last name').value
    assert 'c/o Mama' == browser.getControl('address prefix', index=0).value
    assert '+4901767654321' == browser.getControl('number', index=1).value
    # The last modification date is also displayed:
    assert '<legend>metadata</legend>' in browser.contents
    assert '<span>Modification Date (UTC)</span>' in browser.contents
    # Changed values get persisted when selecting `Apply`:
    browser.getControl('first name').value = u'Peter'
    browser.getControl('birth date').value = '2001 1 1 '
    browser.getControl('zip', index=0).value = u'99991'
    browser.getControl('e-mail address', index=1).value = u'petra9@example.com'
    browser.getControl('URL', index=0).value = 'http://petra23.example.com'
    browser.getControl('number', index=1).value = u'110'
    browser.getControl('Apply').click()
    assert 'Data successfully updated.' == browser.message
    assert browser.PERSONS_LIST_URL == browser.url
    assert (['Tester', 'Peter'] ==
            browser.etree.xpath('//table/tbody/tr/td/a/text()'))
    # The changes show up in the edit form, too:
    browser.open(browser.PERSON_EDIT_URL)
    assert 'Peter' == browser.getControl('first name').value
    assert 'Tester' == browser.getControl('last name').value
    assert '2001 1 1 ' == browser.getControl('birth date').value
    assert '99991' == browser.getControl('zip', index=0).value
    assert 'petra9@example.com' == browser.getControl(
        'e-mail address', index=1).value
    assert ('http://petra23.example.com' ==
            browser.getControl('URL', index=0).value)
    assert '110' == browser.getControl('number', index=1).value


def test_person__PersonEditForm__2(person_data, browser):
    """`PersonEditForm` allows to cancel changes in person's data."""
    browser.login('editor')
    browser.open(browser.PERSON_EDIT_URL)
    # If the user edits and decides to discard his changes he can choose the
    # cancel button and is set back to the address book overview. The
    # changes were not saved:
    browser.getControl('first name').value = u'Hans'
    browser.getControl('number', index=1).value = u'110'
    browser.getControl('Cancel').click()
    assert 'No changes were applied.' == browser.message
    assert browser.PERSONS_LIST_URL == browser.url
    assert (['Tester', 'Petra'] ==
            browser.etree.xpath('//table/tbody/tr/td/a/text()'))
    browser.open(browser.PERSON_EDIT_URL)
    assert 'Petra' == browser.getControl('first name').value
    assert '+4901767654321' == browser.getControl('number', index=1).value


def test_person__PersonEditForm__3(person_data, browser):
    """`PersonEditForm` validates the form input."""
    browser.login('editor')
    browser.open(browser.PERSON_EDIT_URL)
    # If the user edits and decides to save his changes he chooses the apply
    # button. If there are errors in the input data an error message is
    # displayed:
    browser.getControl('birth date').value = u'qqqq'
    browser.getControl('Apply').click()
    assert [] == browser.message
    assert (
        ["The datetime string did not match the pattern u'yyyy MMM d '."] ==
        browser.etree.xpath('//ul[@class="errors"]/li/div/text()'))


def test_person__PersonEditForm__4(person_with_field_data, browser):
    """User defined fields show up on the edit form and contain their value."""
    browser.login('editor')
    browser.open(browser.PERSON_EDIT_URL)
    assert 'my value' == browser.getControl('foobar').value
    browser.getControl('foobar').value = 'monster'
    browser.getControl('Apply').click()
    assert 'Data successfully updated.' == browser.message
    assert browser.PERSONS_LIST_URL == browser.url
    browser.getLink('Tester').click()
    assert 'monster' == browser.getControl('foobar').value
    browser.getControl('Cancel').click()
    assert 'No changes were applied.' == browser.message


def test_person__PersonEditForm__5(person_with_field_data, browser):
    """An `editor` is able to edit content in user definded fields."""
    browser.login('editor')
    browser.open(browser.PERSON_EDIT_URL)
    assert 'my value' == browser.getControl('foobar').value
    browser.getControl('foobar').value = 'editor was here'
    browser.getControl('Apply').click()
    assert 'Data successfully updated.' == browser.message
    assert browser.PERSONS_LIST_URL == browser.url
    # The entered values got persisted:
    browser.getLink('Tester').click()
    assert browser.PERSON_EDIT_URL == browser.url
    assert 'editor was here' == browser.getControl('foobar').value


def test_person__PersonEditForm__6(person_with_field_data, browser):
    """A `visitor` is only able to see content in all kinds of fields.

    There is neither a clone button nor a delete button.

    """
    browser.login('visitor')
    browser.open(browser.PERSON_EDIT_URL)
    assert 'petra@example.com' in browser.contents
    assert 'my value' in browser.contents
    assert ['form.buttons.apply',
            'form.buttons.cancel',
            'form.buttons.export'] == browser.all_control_names


@pytest.mark.parametrize('username', ('editor', 'mgr'))
def test_person__PersonEditForm__7(
        person_with_field_data, browser2, browser, username):
    """A deleted field is not displayed any more in the edit form."""
    delete_field(browser2, 'Field-1')
    browser.login(username)
    browser.open(browser.PERSON_EDIT_URL)
    with pytest.raises(LookupError):
        browser.getControl('foobar')


def test_person__PersonEditForm__8(person_with_field_data, browser2, browser):
    """A deleted field is not displayed any more in the edit form."""
    delete_field(browser2, 'Field-1')
    browser.login('visitor')
    browser.open(browser.PERSON_EDIT_URL)
    assert 'foobar' not in browser.contents


def test_person__PersonEditForm__9(person_data, browser):
    """A keyword can be assigned using `PersonEditForm`."""
    browser.login('editor')
    browser.open(browser.PERSON_EDIT_URL)
    browser.getControl('last name').value = 'keyword test'
    # Keyword assignments are stored when saving:
    assert browser.getControl('church').selected
    assert browser.getControl('family').selected
    browser.getControl('church').click()
    browser.getControl('Apply').click()
    assert 'Data successfully updated.' == browser.message
    address_book = person_data
    family = address_book.keywords.get_keyword_by_title(u'family')
    with zope.component.hooks.site(address_book):
        assert set([family]) == set(address_book['Person'].keywords)


def test_person__PersonEditForm__10(person_data, browser):
    """`PersonEditForm` shows entities and fields in default sort order."""
    browser.login('editor')
    browser.open(browser.PERSON_EDIT_URL)
    assert ['Person_0.first_name',
            'Person_0.last_name',
            'Person_0.birth_date',
            'Person_0.keywords:list',
            'Person_0.notes',
            'defaults.default_postal_address:list',
            'defaults.default_phone_number:list',
            'defaults.default_email_address:list',
            'defaults.default_home_page_address:list',
            'Postaladdress_0.address_prefix',
            'Postaladdress_0.street',
            'Postaladdress_0.city',
            'Postaladdress_0.zip',
            'Postaladdress_0.country:list',
            'Postaladdress_1.address_prefix',
            'Postaladdress_1.street',
            'Postaladdress_1.city',
            'Postaladdress_1.zip',
            'Postaladdress_1.country:list',
            'Phonenumber_0.number',
            'Phonenumber_1.number',
            'Emailaddress_0.email',
            'Emailaddress_1.email',
            'Homepageaddress_0.url',
            'Homepageaddress_1.url',
            'buttons.apply',
            'buttons.clone_person',
            'buttons.delete_person',
            'buttons.delete_entry',
            'buttons.export',
            'buttons.cancel'] == form_fields_names(browser)


def test_person__PersonEditForm__11(person_data, browser):
    """`PersonEditForm` respects the user defined entity  sort order."""
    browser.login('editor')
    person_entity = IEntity(IPerson)
    with zope.component.hooks.site(person_data):
        order = zope.component.getUtility(IEntityOrder)
        # When moving the person one position down, the sort order changes
        # accordingly:
        order.down(person_entity, delta=1)
    browser.open(browser.PERSON_EDIT_URL)
    assert ['defaults.default_postal_address:list',
            'defaults.default_phone_number:list',
            'defaults.default_email_address:list',
            'defaults.default_home_page_address:list',
            'Person_0.first_name',
            'Person_0.last_name',
            'Person_0.birth_date',
            'Person_0.keywords:list',
            'Person_0.notes',
            'Postaladdress_0.address_prefix',
            'Postaladdress_0.street',
            'Postaladdress_0.city',
            'Postaladdress_0.zip',
            'Postaladdress_0.country:list',
            'Postaladdress_1.address_prefix',
            'Postaladdress_1.street',
            'Postaladdress_1.city',
            'Postaladdress_1.zip',
            'Postaladdress_1.country:list',
            'Phonenumber_0.number',
            'Phonenumber_1.number',
            'Emailaddress_0.email',
            'Emailaddress_1.email',
            'Homepageaddress_0.url',
            'Homepageaddress_1.url',
            'buttons.apply',
            'buttons.clone_person',
            'buttons.delete_person',
            'buttons.delete_entry',
            'buttons.export',
            'buttons.cancel'] == form_fields_names(browser)


def test_person__PersonEditForm__12(address_book, UserFactory, browser):
    """In `PersonEditForm` is no delete button for a persons who is a user."""
    UserFactory(address_book, u'Andronicus', u'Loscher', u'l@scher.de',
                u'7zhn8ujm', ['Visitor'])
    browser.login('mgr')
    browser.open(browser.PERSON_EDIT_URL)
    assert ['form.buttons.apply',
            'form.buttons.cancel',
            'form.buttons.clone_person',
            'form.buttons.delete_entry',
            'form.buttons.export'] == browser.submit_control_names


def test_person__PersonEditForm__13(
        address_book, FullPersonFactory, FileFactory, browser):
    """It allows to change the name and mimetype of a file entity."""
    FileFactory(FullPersonFactory(address_book, u'Test'), u'my-file.txt',
                data='boring text')
    browser.login('editor')
    browser.open(browser.PERSON_EDIT_URL)
    browser.getControl('name', index=2).value = 'my nice file.txt'
    browser.getControl('Mime Type').value = 'text/example'
    browser.getControl('Apply').click()
    assert 'Data successfully updated.' == browser.message
    assert browser.PERSONS_LIST_URL == browser.url


def test_person__PersonEditForm__14(
        address_book, FullPersonFactory, FileFactory, browser, tmpfile):
    """It allows to upload a new file.

    This changes the name and the mime type if necessary. When the browser does
    not know the mime type and sends ``application/octet-stream``, the mime
    type is guessed using the file extension and file contents.

    """
    FileFactory(FullPersonFactory(address_book, u'Test'), u'my-file.txt',
                data='boring text')
    browser.login('editor')
    browser.open(browser.PERSON_EDIT_URL)
    fh, filename = tmpfile('special data, blah', '.js')
    browser.getControl('file', index=1).add_file(
        fh, 'application/octet-stream', filename)
    browser.getControl('Apply').click()
    assert 'Data successfully updated.' == browser.message
    assert browser.PERSONS_LIST_URL == browser.url
    # The downloaded file behaves accordingly:
    browser.open(browser.PERSON_EDIT_URL)
    assert browser.getControl('file name').value == filename
    assert 'application/javascript' == browser.getControl('Mime Type').value


def test_person__PersonEditForm__15(
        address_book, FullPersonFactory, FileFactory, browser, tmpfile):
    """It correctly handles a missing mime-type.

    The mime type is optional, if the browser of the client does not send
    a content type and zope.mimetype can't determine the mime type from
    the filename or file content, it is set to ``application/octet-stream``.

    """
    FileFactory(FullPersonFactory(address_book, u'Test'), u'my-file.txt',
                data='boring text')
    browser.login('editor')
    browser.open(browser.PERSON_EDIT_URL)
    fh, filename = tmpfile('Ä, no content type, huh', '')
    browser.getControl('file', index=1).add_file(fh, '', filename)
    browser.getControl('Apply').click()
    assert 'Data successfully updated.' == browser.message
    assert browser.PERSONS_LIST_URL == browser.url
    browser.open(browser.PERSON_EDIT_URL)
    assert 'application/octet-stream' == browser.getControl('Mime Type').value


def test_person__PersonEditForm__16(
        address_book, FieldFactory, FullPersonFactory, FileFactory, browser):
    """The data in the user defined field on a file can be edited."""
    field_name = FieldFactory(address_book, IFile, 'Text', u'mynotes').__name__
    person = FullPersonFactory(address_book, u'Tester')
    FileFactory(person, **{'filename': u'foo.txt', field_name: 'first letter'})
    browser.login('editor')
    browser.open(browser.PERSON_EDIT_URL)
    assert 'first letter' == browser.getControl('mynotes').value
    browser.getControl('mynotes').value = 'second letter'
    browser.getControl('Apply').click()
    assert 'Data successfully updated.' == browser.message
    assert browser.PERSONS_LIST_URL == browser.url
    browser.getLink('Tester').click()
    assert 'second letter' == browser.getControl('mynotes').value


def test_person__PersonEditForm__17(
        address_book, FullPersonFactory, FileFactory, browser):
    """A visitor is not able to edit a file."""
    FileFactory(FullPersonFactory(address_book, u'Test'), u'my-file.txt',
                data='boring text')
    browser.login('visitor')
    browser.open(browser.PERSON_EDIT_URL)
    # There are neither any input widgets nor a delete button:
    assert ['form.buttons.apply',
            'form.buttons.cancel',
            'form.buttons.export'] == browser.all_control_names


def test_person__PersonEditGroup__1(person_data, browser):
    """`PersonEditGroup` respects the user defined field sort order."""
    browser.login('editor')
    postal_address_entity = IEntity(IPostalAddress)
    # When switching the address prefix and the street name of the postal
    # address the person edit form changes accordingly:
    with zope.component.hooks.site(person_data):
        postal_address_entity.setFieldOrder(('street', 'address_prefix'))
    browser.open(browser.PERSON_EDIT_URL)
    assert ['Person_0.first_name',
            'Person_0.last_name',
            'Person_0.birth_date',
            'Person_0.keywords:list',
            'Person_0.notes',
            'defaults.default_postal_address:list',
            'defaults.default_phone_number:list',
            'defaults.default_email_address:list',
            'defaults.default_home_page_address:list',
            'Postaladdress_0.street',
            'Postaladdress_0.address_prefix',
            'Postaladdress_0.city',
            'Postaladdress_0.zip',
            'Postaladdress_0.country:list',
            'Postaladdress_1.street',
            'Postaladdress_1.address_prefix',
            'Postaladdress_1.city',
            'Postaladdress_1.zip',
            'Postaladdress_1.country:list',
            'Phonenumber_0.number',
            'Phonenumber_1.number',
            'Emailaddress_0.email',
            'Emailaddress_1.email',
            'Homepageaddress_0.url',
            'Homepageaddress_1.url',
            'buttons.apply',
            'buttons.clone_person',
            'buttons.delete_person',
            'buttons.delete_entry',
            'buttons.export',
            'buttons.cancel'] == form_fields_names(browser)


def test_person__PersonEditGroup__2(address_book, FullPersonFactory, browser):
    """`PersonEditGroup` stores modification date only for changed entries."""
    # To ease the comparisons in the test we set all modification dates to a
    # default value:
    person = FullPersonFactory(address_book, u'Tester')
    dt = set_modified(person, 2000, 1, 1)
    for entry in person.values():
        set_modified(entry, 2000, 1, 1)
    browser.login('editor')
    browser.open(browser.PERSON_EDIT_URL)
    browser.getControl('city').value = 'Heretown'
    browser.getControl('Apply').click()
    assert 'Data successfully updated.' == browser.message
    assert browser.PERSONS_LIST_URL == browser.url
    assert IZopeDublinCore(person).modified == dt
    assert IZopeDublinCore(person['PhoneNumber']).modified == dt
    assert IZopeDublinCore(person['HomePageAddress']).modified == dt
    assert IZopeDublinCore(person['EMailAddress']).modified == dt
    # Only postal address has a changed modification date:
    assert IZopeDublinCore(person['PostalAddress']).modified != dt


def test_person__PersonEditGroup__3(address_book, FullPersonFactory, browser):
    """`PersonEditGroup` does not alter entity mod-dates if person changed."""
    # To ease the comparisons in the test we set all modification dates to a
    # default value:
    person = FullPersonFactory(address_book, u'Tester')
    dt = set_modified(person, 2000, 1, 1)
    for entry in person.values():
        set_modified(entry, 2000, 1, 1)
    browser.login('editor')
    browser.open(browser.PERSON_EDIT_URL)
    browser.getControl('first name').value = 'Hans'
    browser.getControl('Apply').click()
    assert 'Data successfully updated.' == browser.message
    assert browser.PERSONS_LIST_URL == browser.url
    # Only person has a changed modification date:
    assert IZopeDublinCore(person).modified != dt
    assert IZopeDublinCore(person['PhoneNumber']).modified == dt
    assert IZopeDublinCore(person['HomePageAddress']).modified == dt
    assert IZopeDublinCore(person['EMailAddress']).modified == dt
    assert IZopeDublinCore(person['PostalAddress']).modified == dt


def test_person__ClonePersonForm__1(person_with_field_data, browser):
    """`ClonePersonForm` allows to clone a person and its data."""
    # As user defined fields and keywords may behave unusual, let's create
    # some:
    browser.login('editor')
    browser.open(browser.PERSON_EDIT_URL)
    browser.getControl('Clone person').click()
    assert browser.PERSON_CLONE_URL == browser.url
    browser.getControl('Yes').click()
    assert '"Tester, Petra" cloned.' == browser.message

    # The URL has changed as the cloned Person gets now edited, but the
    # entered data is still the same:
    assert browser.PERSON_EDIT_URL + '-2' == browser.url
    assert 'Tester' == browser.getControl('last name').value
    assert 'Petra' == browser.getControl('first name').value
    assert ['church', 'family'] == browser.getControl('keywords').displayValue
    assert 'my value' == browser.getControl('foobar').value
    assert 'petra@example.com' == browser.getControl('e-mail', index=1).value
    assert '+4901767654321' == browser.getControl('number', index=1).value
    assert ('http://petra.example.com' ==
            browser.getControl('URL', index=0).value)

    # Changing data in the clone does not change the original:
    browser.getControl('last name').value = 'Testerella'
    browser.getControl('first name').value = 'Marta'
    browser.getControl('keywords').displayValue = ['friend']
    browser.getControl('foobar').value = 'other value'
    browser.getControl('e-mail', index=1).value = 'm@test333.de'
    browser.getControl('number', index=1).value = '333'
    browser.getControl('URL', index=0).value = 'http://test333.de'
    browser.getControl('Apply').click()
    assert 'Data successfully updated.' == browser.message
    assert browser.PERSONS_LIST_URL == browser.url
    browser.getLink('Tester').click()
    assert 'Tester' == browser.getControl('last name').value
    assert 'Petra' == browser.getControl('first name').value
    assert ['church', 'family'] == browser.getControl('keywords').displayValue
    assert 'my value' == browser.getControl('foobar').value
    assert 'petra@example.com' == browser.getControl('e-mail', index=1).value
    assert '+4901767654321' == browser.getControl('number', index=1).value
    assert ('http://petra.example.com' ==
            browser.getControl('URL', index=0).value)

    # Changing data in the original does not change the clone:
    browser.getControl('last name').value = 'Testr'
    browser.getControl('keywords').displayValue = ['church', 'friend']
    browser.getControl('foobar').value = 'your value'
    browser.getControl('number', index=1).value = '012354-234234-23'
    browser.getControl('Apply').click()
    assert 'Data successfully updated.' == browser.message
    assert browser.PERSONS_LIST_URL == browser.url
    browser.getLink('Testerella').click()
    assert 'Testerella' == browser.getControl('last name').value
    assert 'Marta' == browser.getControl('first name').value
    assert ['friend'] == browser.getControl('keywords').displayValue
    assert 'other value' == browser.getControl('foobar').value
    assert 'm@test333.de' == browser.getControl('e-mail', index=1).value
    assert '333' == browser.getControl('number', index=1).value
    assert 'http://test333.de' == browser.getControl('URL', index=0).value


def test_person__ClonePersonForm__2(address_book, FullPersonFactory, browser):
    """`ClonePersonForm` cannot be accessed by a visitor."""
    FullPersonFactory(address_book, u'Test')
    browser.login('visitor')
    with pytest.raises(HTTPError) as err:
        browser.open(browser.PERSON_CLONE_URL)
    assert 'HTTP Error 403: Forbidden' == str(err.value)


def test_person__DefaultSelectGroup__1(person_data, browser):
    """`DefaultSelectGroup` allows to change the default addresses."""
    browser.login('editor')
    browser.open(browser.PERSON_EDIT_URL)
    # There is always one main address of each kind (by default the first
    # one created). All addresses of each kind are displayed in a drop-down:
    assert(
        ['c/o Mama, Demoweg 23, 88888, Testhausen, Austria',
         'RST-Software, Forsterstraße 302a, 98344, Erfurt, Germany'] ==
        browser.getControl('main postal address').displayOptions)
    assert (['c/o Mama, Demoweg 23, 88888, Testhausen, Austria'] ==
            browser.getControl('main postal address').displayValue)

    assert (['+4901767654321', '+4901761234567'] ==
            browser.getControl('main phone number').displayOptions)
    assert (['+4901767654321'] ==
            browser.getControl('main phone number').displayValue)

    assert (['petra@example.com', 'pt@rst.example.edu'] ==
            browser.getControl('main e-mail address').displayOptions)
    assert (['petra@example.com'] ==
            browser.getControl('main e-mail address').displayValue)

    assert (['http://petra.example.com',
             'http://www.rst.example.edu'] ==
            browser.getControl('main home page address').displayOptions)
    assert (['http://petra.example.com'] ==
            browser.getControl('main home page address').displayValue)
    # Set the other addresses as default:

    browser.getControl('main postal address').displayValue = [
        'RST-Software, Forsterstra\xc3\x9fe 302a, 98344, Erfurt, Germany']
    browser.getControl('main e-mail address').displayValue = [
        'pt@rst.example.edu']
    browser.getControl('main home page address').displayValue = [
        'http://www.rst.example.edu']
    browser.getControl('Apply').click()
    assert 'Data successfully updated.' == browser.message
    browser.open(browser.PERSON_EDIT_URL)
    assert (['http://www.rst.example.edu'] ==
            browser.getControl('main home page address').displayValue)


def test_person__DefaultSelectGroup__2(person_data, browser):
    """`DefaultSelectGroup` respects the user defined entity sort order."""
    browser.login('editor')
    # The default sort order is tested in test_person__PersonEditForm__10!
    postal_address_entity = IEntity(IPostalAddress)
    with zope.component.hooks.site(person_data):
        order = zope.component.getUtility(IEntityOrder)
        # When moving the postal address one position down, the sort order
        # changes accordingly:
        order.down(postal_address_entity, delta=1)
    browser.open(browser.PERSON_EDIT_URL)
    assert (['default_phone_number',
             'default_postal_address',
             'default_email_address',
             'default_home_page_address'] ==
            [x.replace('defaults.', '').replace(':list', '')
             for x in form_fields_names(browser)
             if x.startswith('defaults.')])


def test_person__DefaultSelectGroup__3(
        translated_address_book, FullPersonFactory, browser):
    """`DefaultSelectGroup` translates country name in the address field."""
    FullPersonFactory(
        translated_address_book, u'Deutsch', postal__city=u'Thun',
        postal__country=gocept.country.db.Country('CH'))
    browser.login('editor')
    browser.lang('de-DE')
    browser.open(browser.PERSON_EDIT_URL)
    assert ['Thun, Schweiz'] == browser.getControl(
        'wichtigste Anschrift').displayValue
    browser.getControl('Abbrechen').click()
    assert u'Keine Änderungen durchgeführt.' == browser.message


def test_person__DefaultSelectGroup__4(
        translated_address_book, FullPersonFactory, browser):
    """`DefaultSelectGroup` uses English country name if requested."""
    FullPersonFactory(
        translated_address_book, u'English', postal__city=u'Thun',
        postal__country=gocept.country.db.Country('CH'))
    browser.login('editor')
    browser.lang('en-US')
    browser.open(browser.PERSON_EDIT_URL)
    assert ['Thun, Switzerland'] == browser.getControl(
        'main postal address').displayValue


def test_person__DeleteSingleEntryForm__1(person_data, browser):
    """It allows to select an entry for delete."""
    browser.login('editor')
    browser.open(browser.PERSON_EDIT_URL)
    browser.getControl('Delete single entry').click()
    assert browser.PERSON_DELETE_ENTRY_URL == browser.url
    assert (
        ['postal address -- c/o Mama, Demoweg 23, 88888, Testhausen, Austria',
         'postal address -- RST-Software, Forsterstra\xc3\x9fe 302a, 98344,'
         ' Erfurt, Germany',
         'phone number -- +4901767654321',
         'phone number -- +4901761234567',
         'e-mail address -- petra@example.com',
         'e-mail address -- pt@rst.example.edu',
         'home page address -- http://petra.example.com',
         'home page address -- http://www.rst.example.edu'] ==
        browser.getControl('Entries').displayOptions)
    # Deletion can be cancelled:
    browser.getControl('Cancel').click()
    assert 'No changes were applied.' == browser.message
    assert browser.PERSON_EDIT_URL == browser.url


def test_person__DeleteSingleEntryForm__2(person_data, browser):
    """It cannot be accessed by a visitor."""
    browser.login('visitor')
    with pytest.raises(HTTPError) as err:
        browser.open(browser.PERSON_DELETE_ENTRY_URL)
    assert 'HTTP Error 403: Forbidden' == str(err.value)


def test_person__DeleteSingleEntryForm__3(person_data, browser):
    """It shows an error message if no entry is selected."""
    browser.login('editor')
    browser.open(browser.PERSON_DELETE_ENTRY_URL)
    browser.getControl('Entries').displayValue = []
    browser.getControl('Delete entry').click()
    assert 'There were some errors.' in browser.contents
    assert 'Required input is missing.' in browser.contents
    assert browser.PERSON_DELETE_ENTRY_URL == browser.url


def test_person__DeletePersonForm__1(person_data, browser):
    """`DeletePersonForm` allows to delete a whole person."""
    browser.login('editor')
    browser.open(browser.PERSON_EDIT_URL)
    browser.getControl('Delete whole person').click()
    assert browser.PERSON_DELETE_URL == browser.url
    browser.getControl('Yes').click()
    assert '"Tester, Petra" deleted.' == browser.message
    assert browser.PERSONS_LIST_URL == browser.url
    browser.open(browser.url)  # simulate reload to get rid of the delete msg
    assert 'Petra' not in browser.contents


def test_person__DeletePersonForm__2(person_data, browser):
    """`DeletePersonForm` can be cancelled."""
    browser.login('editor')
    browser.open(browser.PERSON_DELETE_URL)
    browser.getControl('No, cancel').click()
    assert 'Deletion canceled.' == browser.message
    assert browser.PERSON_EDIT_URL == browser.url


def test_person__DeletePersonForm__3(person_data, browser):
    """`DeletePersonForm` cannot be accessed by a visitor."""
    browser.login('visitor')
    with pytest.raises(HTTPError) as err:
        browser.open(browser.PERSON_DELETE_URL)
    assert 'HTTP Error 403: Forbidden' == str(err.value)


def test_person__DeletePersonForm__4(address_book, UserFactory, browser):
    """Directly accessing `DeletePersonForm` for user leads to an error msg."""
    UserFactory(address_book, u'Andronicus', u'Loscher', u'l@scher.de',
                u'7zhn8ujm', ['Visitor'])
    browser.login('mgr')
    browser.open(browser.PERSON_DELETE_URL)
    browser.getControl('Yes').click()
    assert ('Failed to delete person: This person is referenced. To delete '
            'this person, remove the reference before.' == browser.message)
    assert browser.PERSONS_LIST_URL == browser.url
    assert 'Loscher' in browser.contents



# -*- coding: utf-8 -*-
import pytest
from icemac.addressbook.address import EMailAddress, HomePageAddress
from zope.schema.interfaces import ConstraintNotSatisfied, InvalidURI


@pytest.mark.parametrize('candidate', [
    u'asdfg',
    u'ich@',
    u'ich@goo@le.de',
    u'ich@local',
    u'ich@local..de',
    u'ich@local.de.'], ids=lambda x: str(x))
def test_interfaces__IEMailAddress__email__1(candidate):
    """The constraint on `email` refuses non-valid email addresses."""
    email_address = EMailAddress()
    with pytest.raises(ConstraintNotSatisfied):
        email_address.email = candidate


@pytest.mark.parametrize('candidate', [
    u'ich@example.org',
    u'ich+du@example.org',
    u'ich=du@example.org',
    u'ich_du@example.org',
    u'ich-du@example.org',
    u'ich+du@a.b.c.d.example.org',
    u'ich+du@example.museum',
    u'ich@ex-ample.de',
    u'ich@ex_ample.de',
    # Some examples from RFC 3696 (Note: not all examples from this RFC are
    # supported!):
    u'user+mailbox@example.com',
    u'customer/department=shipping@example.com',
    u'$A12345@example.com',
    u'!def!xyz%abc@example.com',
    u'_somename@example.com'], ids=lambda x: str(x))
def test_interfaces__IEMailAddress__email__2(candidate):
    """The constraint on `email` allows valid email addresses."""
    email_address = EMailAddress()
    email_address.email = candidate


@pytest.mark.parametrize('candidate', [
    'asdfg',
    'www.example.com'])
def test_interfaces__IHomePageAddress__url__1(candidate):
    """The constraint on `url` refuses non-valid URIs."""
    home_page_address = HomePageAddress()
    with pytest.raises(InvalidURI):
        home_page_address.url = candidate


@pytest.mark.parametrize('candidate', [
    'http://www.example.org',
    'http://www2.example.org',
    'http://a.b.c.d.example.org',
    'http://example.museum'])
def test_interfaces__IHomePageAddress__url__2(candidate):
    """The constraint on `url` allows valid URIs."""
    home_page_address = HomePageAddress()
    home_page_address.url = candidate



from zope.testbrowser.browser import HTTPError
import pytest


def test_keyword__Table__1(address_book, browser):
    """The `Table` shows that initially no keywords are defined."""
    browser.login('editor')
    browser.open(browser.MASTER_DATA_URL)
    browser.getLink('Keywords').click()
    assert browser.KEYWORDS_LIST_URL == browser.url
    assert '<div id="content">No keywords defined yet.' in browser.contents


def test_keyword__Table__2(address_book, KeywordFactory, browser):
    """The `Table` shows keywords case insensitive alphabetically sorted."""
    KeywordFactory(address_book, u'Company co-worker')
    KeywordFactory(address_book, u'church')
    KeywordFactory(address_book, u'friend')
    KeywordFactory(address_book, u'family')
    browser.login('editor')
    browser.open(browser.KEYWORDS_LIST_URL)
    assert ['church',
            'Company co-worker',
            'family',
            'friend'] == browser.etree.xpath('//tr/td/a/text()')


def test_keyword__Table__3(address_book, KeywordFactory, browser):
    """A visitor is allowed to see the keywords in the `Table`."""
    KeywordFactory(address_book, u'Arbeit')
    browser.login('visitor')
    browser.open(browser.KEYWORDS_LIST_URL)
    assert ['Arbeit'] == browser.etree.xpath('//tbody/tr/td/a/text()')


def test_keyword__AddForm__1(address_book, browser):
    """The `AddForm` allows to add a new keyword."""
    browser.login('editor')
    browser.open(browser.KEYWORDS_LIST_URL)
    browser.getLink('keyword').click()
    assert browser.KEYWORD_ADD_URL == browser.url
    browser.getControl('keyword').value = 'company coworkr'
    browser.getControl('Add').click()
    assert '"company coworkr" added.' == browser.message


def test_keyword__AddForm__2(address_book, KeywordFactory, browser):
    """`AddForm` prevents from creating duplicate keyword titles."""
    KeywordFactory(address_book, u'Kirche')
    browser.login('editor')
    browser.open(browser.KEYWORD_ADD_URL)
    browser.getControl('keyword').value = 'Kirche'
    browser.getControl('Add').click()
    assert [] == browser.message
    assert browser.KEYWORD_ADD_URL == browser.url
    assert ['This keyword already exists.'] == browser.etree.xpath(
        '/descendant-or-self::div[@class="error"]/text()')


def test_keyword__EditForm__1(address_book, KeywordFactory, browser):
    """The `EditForm` allows to edit the keyword."""
    KeywordFactory(address_book, u'company coworkr')
    browser.login('editor')
    browser.open(browser.KEYWORDS_LIST_URL)
    browser.getLink('company coworkr').click()
    assert browser.KEYWORD_EDIT_URL == browser.url
    assert 'company coworkr' == browser.getControl('keyword').value
    # The last modification date is also displayed:
    assert '<legend>metadata</legend>' in browser.contents
    assert '<span>Modification Date (UTC)</span>' in browser.contents
    # We correct the typo to show the edit form works:
    browser.getControl('keyword').value = 'Company co-worker'
    browser.getControl('Apply').click()
    assert 'Data successfully updated.' == browser.message
    # To show that the changes where saved we go again to the form:
    browser.getLink('Company co-worker').click()
    assert 'Company co-worker' == browser.getControl('keyword').value


def test_keyword__EditForm__2(address_book, KeywordFactory, browser):
    """The user can cancel changes he made in the `EditForm`."""
    KeywordFactory(address_book, u'Company co-worker')
    browser.login('editor')
    browser.open(browser.KEYWORD_EDIT_URL)
    browser.getControl('keyword').value = 'typo'
    browser.getControl('Cancel').click()
    assert 'No changes were applied.' == browser.message
    # The value really did not change:
    browser.getLink('Company co-worker').click()
    assert 'Company co-worker' == browser.getControl('keyword').value


def test_keyword__EditForm__3(
        address_book, KeywordFactory, FullPersonFactory, browser):
    """Changing a keyword in `EditForm` changes the keyword at the person."""
    family = KeywordFactory(address_book, u'family')
    KeywordFactory(address_book, u'church')
    FullPersonFactory(address_book, u'Tester', keywords=set([family]))
    browser.login('editor')
    browser.open(browser.KEYWORD_EDIT_URL)
    browser.getControl('keyword').value = 'Familie'
    browser.getControl('Apply').click()
    assert 'Data successfully updated.' == browser.message

    browser.getLink('church').click()
    browser.getControl('keyword').value = 'Kirche'
    browser.getControl('Apply').click()
    assert 'Data successfully updated.' == browser.message

    browser.getLink('Person list').click()
    browser.getLink('Tester').click()
    assert browser.getControl('Familie').selected
    assert not browser.getControl('Kirche').selected
    with pytest.raises(LookupError) as err:
        browser.getControl('church')
    assert unicode(err.value).startswith("label 'church'")


def test_keyword__EditForm__4(
        address_book, KeywordFactory, FullPersonFactory, browser):
    """`EditForm` does not display a delete button for a referenced keyword."""
    FullPersonFactory(address_book, u'Tester',
                      keywords=set([KeywordFactory(address_book, u'keyword')]))
    browser.login('editor')
    browser.open(browser.KEYWORD_EDIT_URL)
    assert ['form.buttons.apply',
            'form.buttons.cancel'] == browser.submit_control_names


def test_keyword__EditForm__5(
        address_book, KeywordFactory, FullPersonFactory, browser):
    """`EditForm` prevents from duplicate keyword titles."""
    # Modifying the title of a keyword so that it is equal to another one
    # leads also to an error message (We create a new keyword first and edit
    # it afterwards.):
    KeywordFactory(address_book, u'Arbeit')
    KeywordFactory(address_book, u'Company co-worker')
    browser.login('editor')
    browser.open(browser.KEYWORD_EDIT_URL)
    browser.getControl('keyword').value = 'Company co-worker'
    browser.getControl('Apply').click()
    assert [] == browser.message
    assert browser.KEYWORD_EDIT_URL
    assert ['This keyword already exists.'] == browser.etree.xpath(
        '/descendant-or-self::div[@class="error"]/text()')


def test_keyword__EditForm__6(address_book, KeywordFactory, browser):
    """A visitor does not see any form fields in the `EditForm`."""
    KeywordFactory(address_book, u'Arbeit')
    browser.login('visitor')
    browser.open(browser.KEYWORD_EDIT_URL)
    # There is no delete button, too:
    assert ['form.buttons.apply',
            'form.buttons.cancel'] == browser.all_control_names


def test_keyword__DeleteForm__1(address_book, KeywordFactory, browser):
    """Deletion can be canceled in the `DeleteForm`."""
    KeywordFactory(address_book, u'friend')
    browser.login('editor')
    browser.open(browser.KEYWORD_EDIT_URL)
    browser.getControl('Delete').click()
    # There is a confirmation dialog where the user has to decide if he
    # really wants to delete the keyword. If he decides not to delete the
    # entry he is led back to the keyword's edit form:
    assert browser.KEYWORD_DELETE_URL == browser.url
    browser.getControl('No, cancel').click()
    assert 'Deletion canceled.' == browser.message
    assert browser.KEYWORD_EDIT_URL == browser.url


def test_keyword__DeleteForm__2(address_book, KeywordFactory, browser):
    """Deletion can be confirmed in the `DeleteForm`."""
    KeywordFactory(address_book, u'friend')
    browser.login('editor')
    browser.open(browser.KEYWORD_EDIT_URL)
    # Submit buttons when the keyword can be deleted:
    assert ['form.buttons.apply',
            'form.buttons.cancel',
            'form.buttons.delete'] == browser.submit_control_names
    # If he decides to delete the entry he is led back to the keyword list
    # where the keyword is no longer listed:
    browser.getControl('Delete').click()
    browser.getControl('Yes').click()
    assert '"friend" deleted.' == browser.message
    assert browser.KEYWORDS_LIST_URL == browser.url
    # Get rid of the message containing "friend":
    browser.open(browser.KEYWORDS_LIST_URL)
    assert 'friend' not in browser.contents


def test_keyword__DeleteForm__3(
        address_book, KeywordFactory, FullPersonFactory, browser):
    """Deletion of a referenced leads to IntegrityError in the `DeleteForm`."""
    from gocept.reference.interfaces import IntegrityError
    FullPersonFactory(address_book, u'Tester',
                      keywords=set([KeywordFactory(address_book, u'keyword')]))
    browser.login('editor')
    browser.open(browser.KEYWORD_DELETE_URL)
    browser.handleErrors = False  # neded for nicer error display below
    with pytest.raises(IntegrityError) as err:
        browser.getControl('Yes').click()
    assert str(err.value).startswith("Can't delete or move")
    assert str(err.value).endswith("is still being referenced.")


def test_keyword__DeleteForm__4(address_book, KeywordFactory, browser):
    """A visitor is not able to open the `DeleteForm` even he knows the URL."""
    KeywordFactory(address_book, u'work')
    browser.login('visitor')
    with pytest.raises(HTTPError) as err:
        browser.open(browser.KEYWORD_DELETE_URL)
    assert 'HTTP Error 403: Forbidden' == str(err.value)



from icemac.addressbook.interfaces import IEntityOrder, IEntity
from icemac.addressbook.interfaces import IPerson, IPhoneNumber
from icemac.addressbook.testing import delete_field
import pytest
import zope.component


def test_browser__CategoryEditForm__1(address_book, browser):
    """`CategoryEditForm` shows an error message on a too little batch size."""
    # The minimum batch size is 1, otherwise an error message is displayed:
    browser.login('visitor')
    browser.open(browser.PREFS_URL)
    browser.getControl('batch size').value = '0'
    browser.getControl('Apply').click()
    assert [] == browser.message
    assert browser.url.startswith(browser.PREFS_URL)
    assert (['Value is too small'] ==
            browser.etree.xpath('//ul[@class="errors"]/li/div/text()'))


def test_browser__CategoryEditForm__2(address_book, FieldFactory, browser):
    """`CategoryEditForm` renders user defined field names in `order_by`."""
    FieldFactory(address_book, IPerson, u'Bool', u'photo permission?')
    browser.login('visitor')
    browser.open(browser.PREFS_URL)
    assert (
        ['person -- first name',
         'person -- last name',
         'person -- birth date',
         'person -- keywords',
         'person -- notes',
         'person -- photo permission?',  # <-- here is the created field
         'postal address -- address prefix',
         'postal address -- street',
         'postal address -- city',
         'postal address -- zip',
         'postal address -- country',
         'phone number -- number',
         'e-mail address -- e-mail address',
         'home page address -- URL'] ==
        browser.getControl('order by').displayOptions)


def test_browser__CategoryEditForm__3(
        address_book, FieldFactory, browser, browser2):
    """`CategoryEditForm` omits deleted user defined fields in `order_by`."""
    field = FieldFactory(
        address_book, IPerson, u'Bool', u'photo permission?').__name__
    delete_field(browser2, field)
    browser.login('visitor')
    browser.open(browser.PREFS_URL)
    # There is no "photo permission" field here:
    assert (
        ['person -- first name',
         'person -- last name',
         'person -- birth date',
         'person -- keywords',
         'person -- notes',
         'postal address -- address prefix',
         'postal address -- street',
         'postal address -- city',
         'postal address -- zip',
         'postal address -- country',
         'phone number -- number',
         'e-mail address -- e-mail address',
         'home page address -- URL'] ==
        browser.getControl('order by').displayOptions)


def test_browser__CategoryEditForm__4(
        address_book, FieldFactory, browser, browser2):
    """`CategoryEditForm` renders a marker for

    * previously selected but
    * than deleted user defined field in `order_by`.
    """
    field = FieldFactory(
        address_book, IPerson, u'Bool', u'photo permission?').__name__
    browser.login('visitor')
    browser.open(browser.PREFS_URL)
    browser.getControl('order by').displayValue = [
        'person -- photo permission?']
    browser.handleErrors = False
    browser.getControl('Apply').click()
    assert 'Data successfully updated.' == browser.message

    delete_field(browser2, field)
    browser.open(browser.PREFS_URL)
    # There is a place holder for the deleted "photo permission" field, so UI
    # does not break:
    assert (
        ['person -- first name',
         'person -- last name',
         'person -- birth date',
         'person -- keywords',
         'person -- notes',
         'postal address -- address prefix',
         'postal address -- street',
         'postal address -- city',
         'postal address -- zip',
         'postal address -- country',
         'phone number -- number',
         'e-mail address -- e-mail address',
         'home page address -- URL',
         'Missing: IcemacAddressbookPersonPerson###Field-1'] ==
        browser.getControl('order by').displayOptions)
    # The deleted field is still selected:
    assert (['Missing: IcemacAddressbookPersonPerson###Field-1'] ==
            browser.getControl('order by').displayValue)


def test_browser__CategoryEditForm__5(address_book, browser):
    """`CategoryEditForm` allows to cancel editing."""
    browser.login('visitor')
    browser.open(browser.PREFS_URL)
    batch_size = browser.getControl('batch size').value
    browser.getControl('batch size').value = '1000'
    browser.getControl('Cancel').click()
    assert 'No changes were applied.' == browser.message
    assert browser.PERSONS_LIST_URL == browser.url
    browser.open(browser.PREFS_URL)
    assert batch_size == browser.getControl('batch size').value


def test_browser__CategoryEditForm__6(address_book, browser):
    """`CategoryEditForm` stores selected columns."""
    browser.login('visitor')
    browser.open(browser.PREFS_URL)
    browser.in_out_widget_select(
        'form.widgets.columns',
        [browser.getControl('person -- birth date', index=0),
         browser.getControl('person -- first name', index=0),
         browser.getControl('person -- last name', index=0)])
    browser.getControl('Apply').click()
    assert 'Data successfully updated.' == browser.message
    browser.open(browser.PREFS_URL)
    assert (['person -- birth date',
             'person -- first name',
             'person -- last name'] ==
            browser.getControl(name='form.widgets.columns.to').displayOptions)


def test_browser__CategoryEditForm__7(address_book, browser):
    """`CategoryEditForm` is not accessible for anonymous users."""
    browser.open(browser.PREFS_URL)
    assert (
        'http://localhost/ab/@@loginForm.html?camefrom=http%3A%2F%2F'
        'localhost%2Fab%2F%2B%2Bpreferences%2B%2B%2Fab%2F%40%40index.html' ==
        browser.url)


@pytest.mark.parametrize('role', ['mgr', 'editor', 'visitor'])
def test_browser__CategoryEditForm__8(address_book, browser, role):
    """`CategoryEditForm` is accessible for all kinds of logged-in users."""
    browser.login(role)
    browser.open(browser.PREFS_URL)
    # The form is displayed:
    assert (['ascending (A-->Z)', 'descending (Z-->A)'] ==
            browser.getControl('sort direction').displayOptions)


def test_browser__CategoryEditForm__9(address_book, browser, browser2):
    """`CategoryEditForm` stores preferences separately for each user."""
    browser.login('visitor')
    browser.open(browser.PREFS_URL)
    assert (['ascending (A-->Z)'] ==
            browser.getControl('sort direction').displayValue)
    browser.getControl('sort direction').displayValue = ['descending (Z-->A)']
    browser.getControl('Apply').click()
    assert 'Data successfully updated.' == browser.message
    browser.open(browser.PREFS_URL)
    assert (['descending (Z-->A)'] ==
            browser.getControl('sort direction').displayValue)

    browser2.login('editor')
    browser2.open(browser.PREFS_URL)
    assert (['ascending (A-->Z)'] ==
            browser2.getControl('sort direction').displayValue)


def test_browser__CategoryEditForm__10(address_book, browser):
    """The columns field of `CategoryEditForm` uses the default sort order."""
    browser.login('visitor')
    browser.open(browser.PREFS_URL)
    assert ([
        'person -- birth date',
        'person -- keywords',
        'person -- notes',
        'postal address -- address prefix',
        'postal address -- street',
        'postal address -- city',
        'postal address -- zip',
        'postal address -- country',
        'phone number -- number',
        'e-mail address -- e-mail address',
        'home page address -- URL'] ==
        browser.getControl(name='form.widgets.columns.from').displayOptions)


def test_browser__CategoryEditForm__11(address_book, browser):
    """The columns field of CategoryEditForm respects changes in sort order."""
    phone_number_entity = IEntity(IPhoneNumber)
    order = zope.component.getUtility(IEntityOrder)
    # When moving the phone number one position up, the sort order in the
    # columns fields changes accordingly:
    order.up(phone_number_entity)
    browser.login('visitor')
    browser.open(browser.PREFS_URL)
    assert ([
        'person -- birth date',
        'person -- keywords',
        'person -- notes',
        'phone number -- number',
        'postal address -- address prefix',
        'postal address -- street',
        'postal address -- city',
        'postal address -- zip',
        'postal address -- country',
        'e-mail address -- e-mail address',
        'home page address -- URL'] ==
        browser.getControl(name='form.widgets.columns.from').displayOptions)


def test_browser__CategoryEditForm__12(address_book, browser):
    """The orderBy field of `CategoryEditForm` uses the default sort order."""
    browser.login('visitor')
    browser.open(browser.PREFS_URL)
    assert ([
        'person -- first name',
        'person -- last name',
        'person -- birth date',
        'person -- keywords',
        'person -- notes',
        'postal address -- address prefix',
        'postal address -- street',
        'postal address -- city',
        'postal address -- zip',
        'postal address -- country',
        'phone number -- number',
        'e-mail address -- e-mail address',
        'home page address -- URL'] ==
        browser.getControl('order by').displayOptions)


def test_browser__CategoryEditForm__13(address_book, browser):
    """The orderBy field of CategoryEditForm respects changes in sort order."""
    phone_number_entity = IEntity(IPhoneNumber)
    order = zope.component.getUtility(IEntityOrder)
    # When moving the phone number one position down, the sort order in the
    # order by field changes accordingly:
    order.down(phone_number_entity)
    browser.login('visitor')
    browser.open(browser.PREFS_URL)
    assert ([
        'person -- first name',
        'person -- last name',
        'person -- birth date',
        'person -- keywords',
        'person -- notes',
        'postal address -- address prefix',
        'postal address -- street',
        'postal address -- city',
        'postal address -- zip',
        'postal address -- country',
        'e-mail address -- e-mail address',
        'phone number -- number',
        'home page address -- URL'] ==
        browser.getControl('order by').displayOptions)


def test_browser__CategoryEditForm__14(address_book, browser):
    """The time zone field is stored at save."""
    browser.login('visitor')
    browser.open(browser.PREFS_TIMEZONE_URL)
    browser.getControl('Time zone').displayValue = ['Europe/Berlin']
    browser.getControl('Apply').click()
    assert 'Data successfully updated.' == browser.message
    assert ['Europe/Berlin'] == browser.getControl('Time zone').displayValue


def test_browser__PrefGroupEditForm__redirect_to_next_url__1(
        address_book, browser):
    """`PrefGroupEditForm` redirects to self after cancel."""
    browser.login('visitor')
    browser.open(browser.PREFS_TIMEZONE_URL)
    assert ['UTC'] == browser.getControl('Time zone').displayValue
    browser.getControl('Time zone').displayValue = ['Africa/Cairo']
    browser.getControl('Cancel').click()
    # After redirect the original value is restored in the field:
    assert ['UTC'] == browser.getControl('Time zone').displayValue
    assert browser.PREFS_TIMEZONE_URL == browser.url



from zope.testbrowser.browser import HTTPError
import pytest


def test__layout_pt__1(address_book, browser):
    """layout.pt renders title of the address book in the title tag and h1."""
    address_book.title = u'My addresses'
    browser.open(browser.ADDRESS_BOOK_DEFAULT_URL)
    assert 'My addresses' == browser.title
    assert ['My addresses'] == browser.etree.xpath('//h1/span/text()')


def test__layout_pt__2(address_book, browser):
    """layout.pt falls back to the default title if no title is set."""
    address_book.title = u''
    browser.open(browser.ADDRESS_BOOK_DEFAULT_URL)
    assert 'icemac.addressbook' == browser.title
    assert ['icemac.addressbook'] == browser.etree.xpath('//h1/span/text()')


def test__layout_pt__3(address_book, browser):
    """layout.pt renders the default title on the root page."""
    browser.login('mgr')  # need to log in to avoid HTTP-401
    browser.open(browser.ROOT_URL)
    assert 'icemac.addressbook' == browser.title
    assert ['icemac.addressbook'] == browser.etree.xpath('//h1/span/text()')


def test__layout_pt__4(address_book, browser):
    """layout.pt renders the default FavIcon on the root page."""
    browser.login('mgr')  # need to log in to avoid HTTP-401
    browser.open(browser.ROOT_URL)
    assert 'href="/++resource++img/favicon-red.ico"' in browser.contents


def test__layout_pt__5(address_book, browser):
    """layout.pt renders the selected FavIcon inside the address book."""
    address_book.favicon = u'/++resource++img/favicon-green.ico'
    browser.open(browser.ADDRESS_BOOK_DEFAULT_URL)
    assert 'href="/++resource++img/favicon-green.ico"' in browser.contents


def test__layout_pt__6(address_book, browser):
    """layout.pt renders the NotFound page nicely."""
    with pytest.raises(HTTPError) as err:
        browser.open('http://localhost/I-do-not-exist')
    assert 'HTTP Error 404: Not Found' == str(err.value)
    # If this test fails there was an exception during rendering the error
    # page:
    assert ('The page you are trying to access is not available' in
            browser.contents)


def test__layout_pt__7(address_book, browser):
    """layout.pt renders the Unauthorized page nicely."""
    with pytest.raises(HTTPError) as err:
        browser.open(browser.ROOT_URL)
    assert 'HTTP Error 401: Unauthorized' == str(err.value)
    # If this test fails there was an exception during rendering the error
    # page:
    assert 'You are not authorized.' in browser.contents


def test_layout__LoggedInUserViewlet__render__1(
        address_book, UserFactory, browser):
    """It renders a link to the edit form of the personal data of the user."""
    user = UserFactory(address_book, u'Hans', u'Tester', u'th@example.com',
                       u'12345678', ['Visitor'])
    assert '1' == user.__name__
    browser.formlogin('th@example.com', '12345678')
    assert browser.ADDRESS_BOOK_WELCOME_URL == browser.url
    assert 'logged in as' in browser.contents
    browser.getLink('th@example.com').click()
    assert browser.PRINCIPAL_EDIT_URL_1 + '/@@index.html' == browser.url


def test_layout__LoggedInUserViewlet__render__2(address_book, browser):
    """It renders only the name if a global user is logged-in."""
    browser.login('visitor')
    browser.open(browser.ADDRESS_BOOK_WELCOME_URL)
    assert 'logged in as "global visitor"' in browser.contents


def test_layout__LoggedInUserViewlet__render__3(address_book, browser):
    """It does not break outside the address book context."""
    browser.login('mgr')
    browser.open(browser.ROOT_URL)
    assert 'logged in as "Manager"' in browser.contents


def test_layout__LoggedInUserViewlet__render__4(address_book, browser):
    """It does not render if no user is logged-in.."""
    browser.open(browser.ADDRESS_BOOK_DEFAULT_URL)
    assert 'logged in as ' not in browser.contents



import icemac.addressbook.browser.base
import icemac.addressbook.browser.menus.menu
import icemac.addressbook.browser.search.result.handler.base


class Checklist(icemac.addressbook.browser.search.result.handler.base.Base,
                icemac.addressbook.browser.base.BaseView):
    """Check-list with person names."""

    def person_list(self):
        for person in self.persons:
            name = icemac.addressbook.interfaces.IPersonName(person).get_name()
            yield {'name': name,
                   'id': person.__name__}

    def person_count(self):
        return len(self.persons)


checklist = icemac.addressbook.browser.menus.menu.SelectMenuItemOn(
    'person-checklist.html')



from __future__ import unicode_literals
import pytest


@pytest.fixture(scope='function')
def search_menu(address_book, browser, sitemenu):
    """Fixture to test the search menu."""
    browser.login('mgr')  # needed for `test__menu_search_menu__5`.
    return sitemenu(browser, 1, 'Search', browser.SEARCH_URL)


def test__menu_search_menu__1(search_menu):
    """Asserting that the menu with the index 1 is `Search`."""
    search_menu.assert_correct_menu_item_is_tested()


def test__menu_search_menu__2(search_menu):
    """The search menu item is selected on the search overview."""
    assert search_menu.item_selected(search_menu.menu_item_URL)


def test__menu_search_menu__3(search_menu):
    """The search menu is not selected on the person list."""
    assert not search_menu.item_selected(search_menu.browser.PERSONS_LIST_URL)


def test__menu_search_menu__4(search_menu):
    """The search menu is selected on the search view."""
    assert search_menu.item_selected(
        'http://localhost/ab/@@multi_keyword.html')


def test__menu_search_menu__5(search_menu):
    """The search menu is selected on the search result handler view."""
    assert search_menu.item_selected('http://localhost/ab/@@multi-update')



import ConfigParser
import collections
import os.path
import sys


def not_matched_prerequisites():
    """Check whether icemac.addressbook can be installed."""
    if os.path.exists('buildout.cfg'):
        return (
            "ERROR: buildout.cfg already exists.\n"
            "       Please (re-)move the existing one and restart install.")
    if sys.version_info[:2] != (2, 7):
        return ("ERROR: icemac.addressbook currently supports only Python 2.7."
                "\n       But you try to install it using Python %s.%s.%s." % (
                    sys.version_info[:3]))
    return False


class Configurator(object):
    """Configure installation.

    user_config ... path to config file with previously entered user values
                    which take precedence over application defaults.
                    Might be None, so no user values are used.
    """

    def __init__(self, user_config=None, stdin=None):
        self.user_config = user_config
        if stdin is not None:
            self.stdin = stdin
        else:
            self.stdin = sys.stdin

    def __call__(self):
        self.load()
        self.print_intro()
        self.get_global_options()
        self.get_server_options()
        self.get_log_options()
        self.print_additional_packages_intro()
        self.get_additional_packages()
        self.get_migration_options()
        self.create_admin_zcml()
        self.create_buildout_cfg()
        self.store()

    def ask_user(
            self, question, section, option, global_default=None, values=()):
        """Ask the user for a value of section/option and store it in config.

        global_default ... use this value as default value when it is not
                           defined in the global config file
        values ... when set, only this value can be entered
        """
        try:
            default = self.get(section, option)
        except ConfigParser.NoOptionError:
            assert global_default is not None
            default = global_default
        while True:
            print ' %s: [%s] ' % (question, default),
            got = self.stdin.readline().strip()
            print
            if not got:
                got = default
            if not values or got in values:
                break
            else:
                print 'ERROR: %r is not in %r.' % (got, values)
                print 'Please choose a value out of the list.'

        self._conf.set(section, option, got)
        return got

    def get(self, section, key):
        return self._conf.get(section, key)

    def load(self):
        """Load the configuration from file.

        Default configuration is always read and user configuration is
        read when `user_config` is set on instance.

        """
        if (self.user_config is not None and
                not os.path.exists(self.user_config)):
            raise IOError('%r does not exist.' % self.user_config)

        to_read = ['install.default.ini']
        if self.user_config is not None:
            to_read.append(self.user_config)

        # create config
        self._conf = ConfigParser.SafeConfigParser(
            dict_type=collections.OrderedDict)
        self._conf.read(to_read)
        if self.user_config is not None:
            self._conf.set('migration', 'old_instance',
                           os.path.dirname(self.user_config))
        else:
            self._conf.set('migration', 'old_instance', '')

    def print_intro(self):
        print 'Welcome to icemac.addressbook installation'
        print
        print 'Hint: to use the default value (the one in [brackets]), ',
        print 'enter no value.'
        print

    def get_global_options(self):
        self.eggs_dir = self.ask_user(
            'Directory to store python eggs', 'install', 'eggs_dir')

    def get_server_options(self):
        self.admin_login = self.ask_user(
            'Log-in name for the administrator', 'admin', 'login')
        self.admin_passwd = self.ask_user(
            'Password for the administrator', 'admin', 'password')
        self.host = self.ask_user('Hostname', 'server', 'host')
        self.port = self.ask_user('Port number', 'server', 'port')
        self.username = self.ask_user(
            'Username whether process should run as different user otherwise '
            'emtpy', 'server', 'username')

    def get_log_options(self):
        print ' Please choose Log-Handler:'
        print '    Details see',
        print 'http://docs.python.org/library/logging.html#handler-objects'
        handlers = (
            'FileHandler', 'RotatingFileHandler', 'TimedRotatingFileHandler')
        log_handler = None
        log_handler = self.ask_user(
            'Log-Handler, choose between ' + ', '.join(handlers),
            'log', 'handler', values=handlers)
        if log_handler == 'RotatingFileHandler':
            self.log_max_bytes = self.ask_user(
                'Maximum file size before rotating in bytes',
                'log', 'max_size')
        elif log_handler == 'TimedRotatingFileHandler':
            self.log_when = self.ask_user(
                'Type of rotation interval, choose between S, M, H, D, W, '
                'midnight', 'log', 'when',
                values=('S', 'M', 'H', 'D', 'W', 'midnight'))
            self.log_interval = self.ask_user(
                'Rotation interval size', 'log', 'interval')
        if log_handler in ('RotatingFileHandler', 'TimedRotatingFileHandler'):
            self.log_backups = self.ask_user(
                'Number of log file backups', 'log', 'backups')
        self.log_handler = log_handler

    def print_additional_packages_intro(self):
        print ' When you need additional packages (e. g. import readers)'
        print ' enter the package names here separated by a newline.'
        print ' When you are done enter an empty line.'
        print ' Known packages:'
        print '   icemac.ab.importer -- Import of CSV files'
        print '   icemac.ab.importxls -- Import of XLS (Excel) files'
        print '   icemac.ab.calendar -- Calendar using persons in address book'

    def get_additional_packages(self):
        packages = []
        index = 1
        while True:
            package = self.ask_user(
                'Package %s' % index, 'packages', 'package_%s' % index,
                global_default='')
            index += 1
            if not package:
                break
            packages.append(package)
        self.packages = packages

    def get_migration_options(self):
        yes_no = ('yes', 'no')
        self.do_migration = self.ask_user(
            ' Migrate old address book content to new instance', 'migration',
            'do_migration', values=yes_no)
        if self.do_migration == 'no':
            return
        print 'The old address book instance must not run during migration.'
        print 'When it runs as demon process the migration script can stop it',
        print 'otherwise you have to stop it manually.'
        self.stop_server = self.ask_user(
            'Old instance is running as a demon process', 'migration',
            'stop_server', values=yes_no)
        self.start_server = self.ask_user(
            'New instance should be started as a demon process after '
            'migration', 'migration', 'start_server', values=yes_no)

    def create_admin_zcml(self):
        print 'creating admin.zcml ...'
        admin_zcml = file('admin.zcml', 'w')
        admin_zcml.write('\n'.join(
            ('<configure xmlns="http://namespaces.zope.org/zope">',
             '  <principal',
             '    id="icemac.addressbook.global.Administrator"',
             '    title="global administrator"',
             '    login="%s"' % self.admin_login,
             '    password_manager="Plain Text"',
             '    password="%s" />' % self.admin_passwd,
             '  <grant',
             '    role="icemac.addressbook.global.Administrator"',
             '    principal="icemac.addressbook.global.Administrator" />',
             '  <grant',
             '    permission="zope.ManageContent"',
             '    principal="icemac.addressbook.global.Administrator" />',
             '</configure>',
             )))
        admin_zcml.close()

    def create_buildout_cfg(self):
        print 'creating buildout.cfg ...'
        buildout_cfg = ConfigParser.SafeConfigParser(
            dict_type=collections.OrderedDict)
        buildout_cfg.add_section('buildout')
        buildout_cfg.set('buildout', 'extends', 'profiles/prod.cfg')
        buildout_cfg.set('buildout', 'newest', 'true')
        buildout_cfg.set('buildout', 'allow-picked-versions', 'true')
        buildout_cfg.set('buildout', 'eggs-directory', self.eggs_dir)
        buildout_cfg.add_section('deploy.ini')
        buildout_cfg.set('deploy.ini', 'host', self.host)
        buildout_cfg.set('deploy.ini', 'port', self.port)
        if self.username:
            buildout_cfg.add_section('zdaemon.conf')
            buildout_cfg.set('zdaemon.conf', 'user', 'user %s' % self.username)
        log_handler = self.log_handler
        if log_handler == 'FileHandler':
            b_log_handler = log_handler
        else:
            # Argh, all other log handlers live in a subpackage
            b_log_handler = 'handlers.' + log_handler
        buildout_cfg.set('deploy.ini', 'log-handler', b_log_handler)
        log_args = getattr(self, '_log_args_{}'.format(log_handler))
        buildout_cfg.set('deploy.ini', 'log-handler-args', log_args)

        if self.packages:
            buildout_cfg.add_section('site.zcml')
            permissions_zcml = (
                '<include package="' +
                '" />\n<include package="'.join(self.packages) +
                '" />')
            buildout_cfg.set('site.zcml', 'permissions_zcml', permissions_zcml)

        buildout_cfg_file = file('buildout.cfg', 'w')
        buildout_cfg.write(buildout_cfg_file)
        if self.packages:
            # ConfigParser can't write '+=' instead of '='
            eggs = 'eggs += %s\n\n' % '\n      '.join(self.packages)
            buildout_cfg_file.write('[app]\n')
            buildout_cfg_file.write(eggs)
            buildout_cfg_file.write('[test]\n')
            buildout_cfg_file.write(eggs)

        buildout_cfg_file.close()

    def store(self):
        print 'saving config ...'
        user_conf = file('install.user.ini', 'w')
        self._conf.write(user_conf)

    @property
    def _log_args_FileHandler(self):
        return "'a'"

    @property
    def _log_args_RotatingFileHandler(self):
        return ', '.join(("'a'", self.log_max_bytes, self.log_backups))

    @property
    def _log_args_TimedRotatingFileHandler(self):
        log_when = "'%s'" % self.log_when
        return ', '.join((log_when, self.log_interval, self.log_backups))



# -*- coding: utf-8 -*-
import icemac.addressbook.browser.search.result.base


class SimplePersonTable(
        icemac.addressbook.browser.search.result.base.BasePersonTable):
    """Result table to register name column on it (via ZCML)."""

    sortOn = u'table-name-1'


class SimpleExportForm(
        icemac.addressbook.browser.search.result.base.BaseSearchResultForm):
    """Export form showing a table containing names of the found persons."""

    table_class = SimplePersonTable



from zope.testbrowser.browser import HTTPError
import pytest


def test_delete__DeleteForm__1(search_data, UserFactory, browser):
    """`DeleteForm` allows an administrator to delete found persons."""
    address_book = search_data
    # Create a user -- the person of a user cannot be deleted using this search
    # result handler.
    UserFactory(address_book, u'Ben', u'Utzer', u'ben@example.com',
                u'12345678', [], keywords=[u'church'])
    browser.login('mgr')
    browser.keyword_search('church')
    # Only the selected persons get deleted. Deselected persons will not:
    browser.getControl(name='persons:list').getControl(
        value="Person-2").selected = False  # This this the person named "Koch"
    browser.getControl('Apply on selected persons').displayValue = [
        'Delete']
    browser.getControl(name='form.buttons.apply').click()
    # The number of persons for deletion is shown on the question screen:
    # (There are 3 persons with the church keyword in the fixture, one got
    # deselected but there is additionally a newly created user.
    assert ['3'] == browser.etree.xpath(
        '//span[@id="form-widgets-count"]/text()')
    assert browser.SEARCH_DELETE_URL == browser.url
    browser.getControl('Yes').click()
    assert 'Selected persons deleted: 2' == browser.message
    assert browser.PERSONS_LIST_URL == browser.url
    # Only the two non-users got deleted:
    assert 'Koch' in browser.contents
    assert 'Utzer' in browser.contents
    assert 'Liebig' not in browser.contents
    assert 'Velleuer' not in browser.contents


def test_delete__DeleteForm__2(search_data, browser):
    """`DeleteForm` can be canceled."""
    browser.login('mgr')
    browser.keyword_search('church', 'Delete')
    # Seleting the `cancel` button leads to the person list without deleting
    # anybody:
    browser.getControl('No, cancel').click()
    assert 'Deletion canceled.' == browser.message
    assert browser.PERSONS_LIST_URL == browser.url
    assert 'Koch' in browser.contents
    assert 'Liebig' in browser.contents
    assert 'Velleuer' in browser.contents


@pytest.mark.parametrize('role', ['editor', 'visitor'])
def test_delete__DeleteForm__3(search_data, browser, role):
    """`DeleteForm` cannot be accessed by non-admin users."""
    browser.login(role)
    browser.keyword_search('church')
    # There is no delete option which can be applied:
    assert (
        ['XLS export main (Exports person data and main addresses resp. '
         'phone numbers.)',
         'XLS export complete (Exports person data and all addresses resp. '
         'phone numbers.)',
         'E-Mail (Creates a link to send e-mails.)',
         'Names (Comma separated list of person names.)',
         'Checklist (List of person names with check-boxes.)',
         "iCalendar export birthday (Export person's birthdays as "
         ".ics file.)"] ==
        browser.getControl('Apply on selected persons').displayOptions)
    with pytest.raises(HTTPError) as err:
        browser.open(browser.SEARCH_DELETE_URL)
    assert 'HTTP Error 403: Forbidden' == str(err.value)



import icemac.addressbook.metadata.interfaces
import zope.schema.fieldproperty
import persistent
import zope.component
import zope.interface


class EditorMetadataStorage(persistent.Persistent):
    """Storage for editor metadata in annotations."""

    zope.component.adapts(zope.interface.Interface)
    zope.interface.implements(icemac.addressbook.metadata.interfaces.IEditor)
    zope.schema.fieldproperty.createFieldProperties(
        icemac.addressbook.metadata.interfaces.IEditor)


editor_metadata_storage = zope.annotation.factory(
    EditorMetadataStorage, key='icemac.metadata')



from icemac.addressbook.browser.search.result.handler.export.base import (
    BaseExport)
import icemac.addressbook.export.xls.simple


class DefaultsExport(BaseExport):
    """Exporter for default data and addresses."""

    exporter_class = icemac.addressbook.export.xls.simple.DefaultsExport


class CompleteExport(BaseExport):
    """Exporter for all data and addresses."""

    exporter_class = icemac.addressbook.export.xls.simple.CompleteExport



import zope.interface
import zope.schema.interfaces
import zope.component.hooks

import icemac.addressbook.interfaces
import icemac.addressbook.principals.interfaces


@zope.interface.implementer(icemac.addressbook.interfaces.ITitle)
def gocept_country_title(obj):
    """Title for objects from gocept.country."""
    return obj.name


@zope.interface.implementer(icemac.addressbook.interfaces.ITitle)
def obj_dot_title(obj):
    """Title for an `obj` where the title is stored on `title` attribute."""
    return obj.title


@zope.component.adapter(zope.interface.Interface)
@zope.interface.implementer(icemac.addressbook.interfaces.ITitle)
def default_title(obj):
    """Default title adapter which returns `str` representation of `obj`."""
    if isinstance(obj, basestring):
        return obj
    return str(obj)


@zope.component.adapter(zope.interface.Interface)
@zope.interface.implementer(icemac.addressbook.principals.interfaces.IRoot)
def principal_root(principal):
    """Return the site root."""
    return zope.component.hooks.getSite()



import icemac.addressbook.browser.base
import icemac.addressbook.browser.interfaces
import icemac.addressbook.interfaces
import z3c.pagelet.browser
import zope.interface
import zope.size.interfaces


@zope.interface.implementer(
    icemac.addressbook.browser.interfaces.IAddressBookBackground)
class FrontPage(z3c.pagelet.browser.BrowserPagelet,
                icemac.addressbook.browser.base.BaseView):
    """Pagelet for the front page."""

    def getAddressBooks(self):
        result = []
        for ab in self.context.values():
            if not icemac.addressbook.interfaces.IAddressBook.providedBy(ab):
                # only show address books
                continue
            result.append(dict(
                title=ab.title,
                url=self.url(ab),
                delete_url=self.url(ab, 'delete-address_book.html'),
                count=zope.size.interfaces.ISized(ab).sizeForDisplay()))
        return result



# -*- coding: utf-8 -*-
from icemac.addressbook.i18n import _
import icemac.addressbook.browser.base
import icemac.addressbook.browser.search.result.handler.manager
import z3c.form.button
import z3c.table.column
import zope.browserpage.viewpagetemplatefile
import zope.interface
import zope.schema


class ISearchResultHanderChoice(zope.interface.Interface):
    """Drop down to select search result handler."""

    search_result_handler = zope.schema.Choice(
        title=_(u'Apply on selected persons'),
        source=icemac.addressbook.browser.search.result.handler.manager.source)


class BaseSearchResultForm(icemac.addressbook.browser.base._AbstractEditForm):
    """Base form of search results allowing handling of results."""

    ignoreContext = True
    interface = ISearchResultHanderChoice
    template = zope.browserpage.viewpagetemplatefile.ViewPageTemplateFile(
        'result.pt')
    id = 'search-export-form'
    table_class = NotImplemented

    def renderTable(self):
        table = self.table_class(self.__parent__, self.request)
        table.update()
        return table.render()

    @z3c.form.button.buttonAndHandler(_('Apply'), name='apply')
    def callHandler(self, action):
        """Call the selected handler after storing data in session."""
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return

        self.session['person_ids'] = self.request.form.get('persons', ())

        # Fanstatic seems to have a problem with cookies and relative
        # redirects (a new cookie is set in this case), so we have to
        # redirect using an absolute URL.
        self.request.response.redirect(self.url(
            self.context, data['search_result_handler'].viewName))


class BasePersonTable(icemac.addressbook.browser.table.Table):
    """Base result table displaying at least a checkbox column (using ZCML)."""

    def update(self):
        self.result = self.context.result
        super(BasePersonTable, self).update()

    @property
    def values(self):
        return self.result


class CheckBoxColumn(z3c.table.column.CheckBoxColumn):
    """Table column to show checkboxes to select result rows for export."""

    header = u'<input type="checkbox" class="checkall" checked="checked" />'
    weight = -1

    def getItemKey(self, item):
        return 'persons:list'

    def isSelected(self, item):
        # The result is not persisted, so always use the default:
        return True



# -*- coding: utf-8 -*-
import transaction
from ..list import PersonList
from icemac.addressbook.fieldsource import tokenize
from icemac.addressbook.interfaces import IEMailAddress
from icemac.addressbook.interfaces import IHomePageAddress, IEntity
from icemac.addressbook.interfaces import IPerson, IPostalAddress, IPhoneNumber
from mock import Mock, patch, PropertyMock
import contextlib
import datetime
import gocept.country.db
import icemac.addressbook.testing
import pytest
import pytz
import zope.component.hooks
import zope.interface


# ready to use fixtures

@pytest.yield_fixture(scope='function')
def MockPersons_22():
    """Create 22 mock persons.

    Also installs patches, so these mock persons can be rendered in PersonList.
    """
    with patch.object(PersonList, 'values',
                      new_callable=PropertyMock) as values, \
            patch('z3c.table.column.absoluteURL') as absoluteURL:
        values.return_value = [
            DummyPerson(
                first_name='first name {}'.format(x), last_name='last name')
            for x in range(22)]
        absoluteURL.return_value = 'http://url.to/person'
        yield


@pytest.yield_fixture(scope='function')
def some_persons(somePersonsS):
    """Some persons to test the sort order."""
    for connection in icemac.addressbook.conftest.pyTestStackDemoStorage(
            somePersonsS.zodb, 'some_persons'):
        yield


# Test helper functions


def assert_person_list(
        browser, result, columns, xpath='//td/text() | //td/a/text()', **kw):
    """Assert the contents of `PersonList` for a given set of columns."""
    prefs_mock = _get_prefs_mock(*columns, **kw)
    browser.login('visitor')
    with patch.object(PersonList, 'prefs',
                      new_callable=PropertyMock) as prefs:
        prefs.return_value = prefs_mock
        browser.open(browser.PERSONS_LIST_URL)
        assert result == browser.etree.xpath(xpath)


@contextlib.contextmanager
def batchSize(size):
    """Context manager to patch the user definable person list batch size."""
    with patch.object(PersonList, 'batchSize', new_callable=PropertyMock) as b:
        b.return_value = size
        yield


def _get_prefs_mock(*columns, **kw):
    """Get a mock for the preferences.

    columns ... show these columns in person lists
                has to be tuples (<entity interface>, <field name>)
    **kw might have the following keys:
        order_by ... order the columns in person lists by this tuple of
                     (<entity interface>, <field name>)
                     Default: (IPerson, 'last_name')
    """
    prefs_mock = Mock()
    prefs_mock.personListTab.batch_size = 20
    prefs_mock.personLists.sort_direction = 'ascending'
    prefs_mock.personLists.columns = [
        tokenize(IEntity(iface), field_name)
        for iface, field_name in columns]
    order_entitiy, order_field = kw.pop('order_by', (IPerson, 'last_name'))
    prefs_mock.personLists.order_by = tokenize(
        IEntity(order_entitiy), order_field)
    if 'sort_direction' in kw:
        prefs_mock.personLists.sort_direction = kw['sort_direction']
    return prefs_mock


# Infrastructure fixtures


@pytest.yield_fixture(scope='session')
def somePersonsS(addressBookS, FullPersonFactory):
    """Create some persons used in tests."""
    for connection in icemac.addressbook.conftest.pyTestStackDemoStorage(
            addressBookS, 'somePersonsS'):
        for address_book in icemac.addressbook.conftest.site(connection):
            FullPersonFactory(address_book, u'Tester', first_name=u'Hans',
                              birth_date=datetime.date(1975, 4, 2),
                              keywords=[u'church', u'Friends', u'family'])
            FullPersonFactory(address_book, u'Streber', first_name=u'Gunter',
                              keywords=[u'Friends'])
            FullPersonFactory(address_book, u'Utzer',
                              birth_date=datetime.date(1974, 5, 14))
            transaction.commit()
        yield connection


class DummyPerson(object):
    """Light weight dummy IPerson object."""

    zope.interface.implements(IPerson)

    def __init__(self, **kw):
        self.__dict__.update(kw)


def test_list__PersonList__1(address_book, browser):
    """The person list is initially empty and shows a message."""
    browser.login('visitor')
    browser.open(browser.ADDRESS_BOOK_DEFAULT_URL)
    browser.getLink('Person list').click()
    assert browser.PERSONS_LIST_URL == browser.url
    assert ('There are no persons entered yet, click on "Add person" to '
            'create one.' in browser.contents)


def test_list__PersonList__2(address_book, browser):
    """A user with the role `visitor` is not allowed to add new persons."""
    from zope.testbrowser.browser import LinkNotFoundError
    browser.login('visitor')
    browser.open(browser.PERSONS_LIST_URL)
    with pytest.raises(LinkNotFoundError):
        browser.getLink('person')


def test_list__PersonList__3(address_book, MockPersons_22, browser):
    """`PersonList` renders a pagination if number of persons > batch size."""
    browser.login('visitor')
    with batchSize(20):
        browser.open(browser.PERSONS_LIST_URL)
    assert ([
        ('1', '/person-list.html?table-batchSize=20&table-batchStart=0'),
        ('2', '/person-list.html?table-batchSize=20&table-batchStart=20')] ==
        [(x.text,
          x.attrib['href'].split(browser.ADDRESS_BOOK_DEFAULT_URL)[-1])
         for x in browser.etree.xpath('//div[@class="batch"]/a')])


def test_list__PersonList__4(address_book, MockPersons_22, browser):
    """`PersonList` renders no pagination if number of persons < batch size."""
    browser.login('visitor')
    with batchSize(23):
        browser.open(browser.PERSONS_LIST_URL)
    assert [] == browser.etree.xpath('//div[@class="batch"]/a')


def test_list__PersonList__5(address_book, MockPersons_22, browser):
    """`PersonList` renders a batch spacer if there are too many batches."""
    browser.login('visitor')
    with batchSize(1):
        browser.open(browser.PERSONS_LIST_URL)
    assert [u'…'] == [
        x.strip()
        for x in browser.etree.xpath('//div[@class="batch"]/text()')
        if x.strip()]


@pytest.mark.parametrize('batch_size, batch_number', ((20, 2), (8, 3)))
def test_list__PersonList__6i(
        address_book, MockPersons_22, browser, batch_size, batch_number):
    """`PersonList` respects the batch size set in the preferences."""
    browser.login('visitor')
    browser.open(browser.PREFS_URL)
    browser.getControl('batch size').value = str(batch_size)
    browser.getControl('Apply').click()
    assert 'Data successfully updated.' == browser.message
    browser.open(browser.PERSONS_LIST_URL)
    assert ([str(x + 1) for x in range(batch_number)] ==
            browser.etree.xpath('//div[@class="batch"]/a/text()'))


def test_list__PersonList__7i(
        address_book, FieldFactory, FullPersonFactory, browser):
    """`PersonList` renders user defined fields."""
    photo_permission = FieldFactory(
        address_book, IPerson, u'Bool', u'photo permission?').__name__
    last_seen = FieldFactory(
        address_book, IPerson, u'Datetime', u'last seen').__name__
    state = FieldFactory(
        address_book, IPostalAddress, u'Choice', u'state',
        values=[u'Sachsen', u'Sachsen-Anhalt', u'Brandenburg']).__name__
    number_of_letters = FieldFactory(
        address_book, IPostalAddress, u'Int', u'number of letters').__name__
    last_time_verified = FieldFactory(
        address_book, IPostalAddress, u'Date', u'last time verified').__name__
    cost_per_minute = FieldFactory(
        address_book, IPhoneNumber, u'Decimal', u'cost per minute').__name__
    mail_box_text = FieldFactory(
        address_book, IPhoneNumber, u'Text', u'mail box text').__name__
    company_home_page = FieldFactory(
        address_book, IHomePageAddress, u'URI', u'company home page').__name__
    home_page_provider = FieldFactory(
        address_book, IHomePageAddress, u'TextLine',
        u'home page provider').__name__

    # Using two persons here to show that user defined fields render
    # an empty value if not set.
    FullPersonFactory(
        address_book, u'Tester', **{
            'person__' + photo_permission: True,
            'postal__' + state: 'Brandenburg',
            'postal__' + number_of_letters: 42,
            'phone__' + cost_per_minute: 1.45,
            'phone__' + mail_box_text: 'Hi, I do not like the boring texts of '
                                       'the answering machines ...',
            'homepage__' + company_home_page: 'http://www.mycompany.de',
            'homepage__' + home_page_provider: 'mycomp',
        })
    FullPersonFactory(
        address_book, u'Utzer', **{
            'postal__' + last_time_verified: datetime.date(2001, 1, 8),
            'person__' + last_seen: datetime.datetime(
                2010, 9, 8, 20, 15, tzinfo=pytz.utc),
        })

    assert_person_list(
        browser,
        ['yes',
         '42',
         'Brandenburg',
         '1.45',
         u'Hi, I do not like the \u2026',
         'http://www.mycompany.de',
         'mycomp',
         'Tester',
         '2010 9 8  20:15:00 ',
         '2001 1 8 ',
         'Utzer'],
        ((IPerson, photo_permission),
         (IPerson, last_seen),
         (IPostalAddress, number_of_letters),
         (IPostalAddress, last_time_verified),
         (IPostalAddress, state),
         (IPhoneNumber, cost_per_minute),
         (IPhoneNumber, mail_box_text),
         (IHomePageAddress, company_home_page),
         (IHomePageAddress, home_page_provider),
         (IPerson, 'last_name')))


def test_list__PersonList__8i(
        address_book, FieldFactory, FullPersonFactory, browser, browser2):
    """`PersonList` omits deleted user defined field selected for display."""
    photo_permission = FieldFactory(
        address_book, IPerson, u'Bool', u'photo permission?').__name__
    FullPersonFactory(address_book, u'Tester', **{photo_permission: False})

    columns = ((IPerson, photo_permission),
               (IPerson, 'last_name'))

    # Column is displayed
    assert_person_list(browser, ['no', 'Tester'], columns)
    # Only a manager can delete the field:
    icemac.addressbook.testing.delete_field(browser2, photo_permission)
    # The delete column gets omitted from person list:
    browser.reload()
    assert_person_list(browser, ['Tester'], columns)


def test_list__PersonList__9i(
        address_book, FieldFactory, FullPersonFactory, browser, browser2):
    """`PersonList` doesn't break when the order-by field gets deleted."""
    photo_permission = FieldFactory(
        address_book, IPerson, u'Bool', u'photo permission?').__name__
    FullPersonFactory(address_book, u'Tester', **{photo_permission: False})
    columns = ((IPerson, photo_permission),
               (IPerson, 'last_name'))
    order_by = (IPerson, photo_permission)

    # Column is displayed
    assert_person_list(browser, ['no', 'Tester'], columns, order_by=order_by)
    # Only a manager can delete the field:
    icemac.addressbook.testing.delete_field(browser2, photo_permission)
    # The delete column gets omitted from person list:
    browser.reload()
    assert_person_list(browser, ['Tester'], columns, order_by=order_by)


def test_list__PersonList__10i(address_book, PersonFactory, browser):
    """`PersonList` respects the columns selected in the preferences."""
    PersonFactory(address_book, u'Tester', notes=u'my notes')
    browser.login('visitor')
    browser.open(browser.PREFS_URL)
    browser.in_out_widget_select(
        'form.widgets.columns',
        [browser.getControl('person -- notes', index=0),
         browser.getControl('person -- last name', index=0)])
    browser.getControl('Apply').click()
    assert 'Data successfully updated.' == browser.message
    browser.open(browser.PERSONS_LIST_URL)
    assert (['my notes', 'Tester'] ==
            browser.etree.xpath('//td/text() | //td/a/text()'))


def test_list__PersonList__11i(
        address_book, FullPersonFactory, EMailAddressFactory, browser):
    """`PersonList` renders all kinds of predefined fields."""
    # Using two persons here to show that the empty value is rendered
    # correctly, too.
    FullPersonFactory(
        address_book, u'Tester', **{
            'person__first_name': u'Hans',
            'person__notes': u'Gave him 20 Dollars in Jan 2010.',
            'person__birth_date': datetime.date(1977, 11, 24),
            'person__keywords': [u'Friend', u'church'],
            'postal__country': gocept.country.db.Country('JP'),
            'phone__number': u'110',
            'homepage__url': 'http://www.mycompany.de',
            'email__email': u'tester@example.com',
        })
    # Country with no value is not displayed.
    utzer = FullPersonFactory(
        address_book, u'Utzer', postal__country=None)
    # This e-mail address will not show up in the list as only the default ones
    # are displayed there:
    EMailAddressFactory(utzer, email=u'ben@utzer.org', set_as_default=False)

    assert_person_list(
        browser,
        ['Hans',
         'Tester',
         u'Gave him 20 Dollars in …',  # notes: truncated + ellipsed
         '1977 11 24 ',
         'church, Friend',  # keywords: comma separated + lower case sorted
         'Japan',
         '110',
         'http://www.mycompany.de',
         'tester@example.com',
         'Utzer'],
        ((IPerson, 'first_name'),
         (IPerson, 'last_name'),
         (IPerson, 'notes'),
         (IPerson, 'birth_date'),
         (IPerson, 'keywords'),
         (IPostalAddress, 'country'),
         (IPhoneNumber, 'number'),
         (IHomePageAddress, 'url'),
         (IEMailAddress, 'email')))


def test_list__PersonList__12i(address_book, FullPersonFactory, browser):
    """`PersonList` renders homepage addresses and emails as links."""
    FullPersonFactory(
        address_book, u'Tester', email__email=u'tester@example.com',
        homepage__url='http://tester.example.com')

    assert_person_list(
        browser,
        ['http://localhost/ab/Person',
         'mailto:tester@example.com',
         'http://tester.example.com'],
        ((IPerson, 'last_name'),
         (IEMailAddress, 'email'),
         (IHomePageAddress, 'url')), xpath='//td/a/@href')


def test_list__PersonList__13i(
        address_book, FullPersonFactory, EMailAddressFactory, browser):
    """`PersonList` respects changed default addresses."""
    tester = FullPersonFactory(
        address_book, u'Tester', email__email=u'1st@example.com')
    email = EMailAddressFactory(
        tester, email=u'2nd@example.com', set_as_default=False)

    assert_person_list(
        browser,
        ['Tester', '1st@example.com'],
        ((IPerson, 'last_name'), (IEMailAddress, 'email')))

    with zope.component.hooks.site(address_book):
        tester.default_email_address = email

    assert_person_list(
        browser,
        ['Tester', '2nd@example.com'],
        ((IPerson, 'last_name'), (IEMailAddress, 'email')))


def test_list__PersonList__14i(some_persons, browser):
    """`PersonList` can be sorted by person's first name."""
    assert_person_list(
        browser,
        ['Utzer', 'Streber', 'Gunter', 'Tester', 'Hans'],
        [(IPerson, 'last_name'),
         (IPerson, 'first_name')],
        order_by=(IPerson, 'first_name'))


def test_list__PersonList__15i(some_persons, browser):
    """`PersonList` sorts by 1st column if order_by col is not displayed.

    Empty values in a column are sorted to the end of the column.
    """
    assert_person_list(
        browser,
        ['1974 5 14 ', 'Utzer', '1975 4 2 ', 'Tester', 'Streber'],
        [(IPerson, 'birth_date'),
         (IPerson, 'last_name')],
        order_by=(IPostalAddress, 'city'))


def test_list__PersonList__16i(some_persons, browser):
    """`PersonList` sorts keyword column by lower case of the cell contents."""
    assert_person_list(
        browser,
        ['Utzer', 'Tester', 'church, family, Friends', 'Streber', 'Friends'],
        [(IPerson, 'last_name'),
         (IPerson, 'keywords')],
        order_by=(IPerson, 'keywords'))


def test_list__PersonList__17i(some_persons, browser):
    """`PersonList` sorts keyword column by lower case of the cell contents."""
    assert_person_list(
        browser,
        ['Utzer', 'Tester', 'church, family, Friends', 'Streber', 'Friends'],
        [(IPerson, 'last_name'),
         (IPerson, 'keywords')],
        order_by=(IPerson, 'keywords'))


def test_list__PersonList__18i(some_persons, browser):
    """`PersonList` allows the user to change the sort order."""
    xpath = '//td/a/text()'
    assert_person_list(
        browser,
        ['Streber', 'Gunter', 'Tester', 'Hans', 'Utzer'],
        [(IPerson, 'last_name'),
         (IPerson, 'first_name')],
        order_by=(IPerson, 'last_name'),
        xpath=xpath)
    browser.getLink('first name').click()
    assert ('http://localhost/ab/@@person-list.html?'
            'table-sortOrder=ascending&table-sortOn=table-first_name-1' ==
            browser.url)
    assert (['Utzer', 'Streber', 'Gunter', 'Tester', 'Hans'] ==
            browser.etree.xpath(xpath))


def test_list__PersonList__19i(some_persons, browser):
    """`PersonList` uses sort direction defined in prefs.

    It additionally allows the user to change the sort direction using a link.
    """
    xpath = '//td/a/text()'
    assert_person_list(
        browser,
        ['Utzer', 'Tester', 'Hans', 'Streber', 'Gunter'],
        [(IPerson, 'last_name'),
         (IPerson, 'first_name')],
        order_by=(IPerson, 'last_name'),
        sort_direction='descending',
        xpath=xpath)
    browser.getLink('last name').click()
    assert ('http://localhost/ab/@@person-list.html?'
            'table-sortOrder=ascending&table-sortOn=table-last_name-0' ==
            browser.url)
    assert (['Streber', 'Gunter', 'Tester', 'Hans', 'Utzer'] ==
            browser.etree.xpath(xpath))



import icemac.addressbook.generations.utils
import icemac.addressbook.interfaces
import zope.component


@icemac.addressbook.generations.utils.evolve_addressbooks
def evolve(addressbook):
    """Create person defaults entity and add it to sort order."""
    person_defaults = icemac.addressbook.interfaces.IEntity(
        icemac.addressbook.interfaces.IPersonDefaults)
    order_storage = zope.component.getUtility(
        icemac.addressbook.interfaces.IOrderStorage)
    order_storage.add(
        person_defaults.name, icemac.addressbook.interfaces.ENTITIES)
    person_defaults_pos = order_storage.get(
        person_defaults.name, icemac.addressbook.interfaces.ENTITIES)
    delta = person_defaults_pos - 3
    order_storage.up(
        person_defaults.name, icemac.addressbook.interfaces.ENTITIES, delta)



# -*- coding: latin-1 -*-
from icemac.addressbook.i18n import _
import collections
import gocept.country
import gocept.country.db
import gocept.reference.field
import icemac.addressbook.sources
import pytz
import re
import zc.sourcefactory.basic
import zc.sourcefactory.contextual
import zc.sourcefactory.source
import zope.component
import zope.interface
import zope.schema


PACKAGE_ID = 'icemac.addressbook'
ENTITIES = 'entites_namespace'
FIELD_NS_PREFIX = 'fields-'
DEFAULT_FAVICON = '/++resource++img/favicon-red.ico'
DEFAULT_STARTPAGE_DATA = (
    'icemac.addressbook.interfaces.IAddressBook', 'welcome.html')


class ITitle(zope.interface.Interface):
    """Title of an entity."""

    def __str__():
        """Return the title of the entity."""


class IImageSource(zope.interface.Interface):
    """Marker interface for a source which uses images as titles.

    Title needs to be the URL of the image.

    """


class IFaviconData(zope.interface.Interface):
    """Data of a favicon."""

    path = zope.interface.Attribute('Path to be used in URL.')
    preview_path = zope.interface.Attribute('Path to be used in preview.')


class FaviconSource(icemac.addressbook.sources.TitleMappingSource):
    """Source containing possbile favicons."""

    class source_class(zc.sourcefactory.source.FactoredSource):
        """We need the interface to register a widget for it."""

        zope.interface.implements(IImageSource)

    @property
    def _mapping(self):
        data = [(x.path, x.preview_path)
                for x in zope.component.subscribers((self, ), IFaviconData)]
        if not data:
            # We are at import time, so the subscribers are not yet set up
            # but IAddressBook.favicon checks if the default value is valid:
            data = [(DEFAULT_FAVICON, DEFAULT_FAVICON)]
        return collections.OrderedDict(sorted(data))


favicon_source = FaviconSource()


class IStartpageData(zope.interface.Interface):
    """Data of a possible startpage after login."""

    iface_name = zope.interface.Attribute(
        'Dotted name to an interface to get the object those URL should '
        'be computed.')
    view = zope.interface.Attribute(
        'View to be displayed on object, might be `None`.')
    title = zope.interface.Attribute('i18n message id for display in UI')


class StartpageSource(icemac.addressbook.sources.TitleMappingSource):
    """Source containing possbile startpages to be shown after login."""

    @property
    def _mapping(self):
        data = [((x.iface_name, x.view), x.title)
                for x in zope.component.subscribers(
                    (self, ), IStartpageData)]
        if not data:
            # We are at import time, so the subscribers are not yet set up
            # but IAddressBook.startpage checks if the default value is
            # valid:
            data = [(DEFAULT_STARTPAGE_DATA, u'')]
        return collections.OrderedDict(sorted(data))

    def getToken(self, value):
        return str(value)


startpage_source = StartpageSource()


class TimeZones(icemac.addressbook.sources.TitleMappingSource):
    """Source of all available time zones."""

    _mapping = collections.OrderedDict(((x, x) for x in pytz.common_timezones))


time_zones = TimeZones()


class IAddressBook(zope.interface.Interface):
    """An address book."""

    keywords = zope.interface.Attribute(
        u'icemac.addressbook.interfaces.IKeywords')
    principals = zope.interface.Attribute(
        u'zope.pluggableauth.plugins.principalfolder.'
        u'IInternalPrincipalContainer')
    importer = zope.interface.Attribute(
        u'icemac.ab.importer.interfaces.IImportContainer')
    entities = zope.interface.Attribute(
        u'icemac.addressbook.interfaces.IEntities')
    orders = zope.interface.Attribute(
        u'icemac.addressbook.interfaces.IOrderStorage')

    title = zope.schema.TextLine(title=_(u'title'))
    favicon = zope.schema.Choice(
        title=_('favicon'),
        source=favicon_source,
        default=DEFAULT_FAVICON)
    startpage = zope.schema.Choice(
        title=_('start page after log-in'),
        source=startpage_source,
        default=DEFAULT_STARTPAGE_DATA)

    time_zone = zope.schema.Choice(
        title=_('Time zone'),
        description=_('Fallback in case a user has not set up his personal '
                      'time zone in the preferences.'),
        source=time_zones, default='UTC')


class IKeywords(zope.interface.Interface):
    """Collection of keywords."""

    def get_keywords():
        """Get an iterable of the keywords in the collection."""

    def get_keyword_by_title(title, default=None):
        """Get the keyword with the given title or `default`."""


class IKeywordTitles(zope.interface.Interface):
    """Collection of keywords titles."""

    def get_titles():
        """Get the titles of the keywords in the collection."""


class IKeyword(zope.interface.Interface):
    """A keyword."""

    title = zope.schema.TextLine(title=_(u'keyword'))


class KeywordSource(zc.sourcefactory.basic.BasicSourceFactory):
    """Source of all keywords in the address book."""

    def getValues(self):
        keywords = zope.component.getUtility(IKeywords)
        return sorted(keywords.get_keywords(), key=lambda x: x.title.lower())

    def getTitle(self, value):
        return value.title


keyword_source = KeywordSource()


class ContextByInterfaceSource(
        zc.sourcefactory.contextual.BasicContextualSourceFactory):
    """Source to select objects from context container by a given interface."""

    def __init__(self, interface):
        super(ContextByInterfaceSource, self).__init__()
        self.interface = interface

    def getValues(self, context):
        for value in context.values():
            if self.interface.providedBy(value):
                yield value

    def getTitle(self, context, value):
        return ITitle(value)


class IPersonName(zope.interface.Interface):
    """Name of a person."""

    first_name = zope.schema.TextLine(title=_(u'first name'), required=False)
    last_name = zope.schema.TextLine(title=_(u'last name'))

    def get_name():
        """Return first name and last name of the person.

        Names are separated by a space.
        """


class IPersonData(zope.interface.Interface):
    """Data of a person."""

    birth_date = zope.schema.Date(title=_(u'birth date'), required=False)
    keywords = gocept.reference.field.Set(
        title=_('keywords'), required=False,
        value_type=zope.schema.Choice(
            title=_('keywords'),
            source=keyword_source))
    notes = zope.schema.Text(title=_(u'notes'), required=False)


class IPerson(IPersonName, IPersonData):
    """A person."""


class IPersonEntity(zope.interface.Interface):
    """Entity of a person."""


class IPostalAddress(IPersonEntity):
    """A postal address."""

    address_prefix = zope.schema.TextLine(
        title=_(u'address prefix'), required=False,
        description=_(u'e. g. company name or c/o'))
    street = zope.schema.TextLine(title=_(u'street'), required=False)
    city = zope.schema.TextLine(title=_(u'city'), required=False)
    zip = zope.schema.TextLine(title=_(u'zip'), required=False)
    country = zope.schema.Choice(
        title=_(u'country'), source=gocept.country.countries,
        required=False, default=gocept.country.db.Country('DE'))


class IEMailAddress(IPersonEntity):
    """An e-mail address."""

    email = zope.schema.TextLine(
        title=_(u'e-mail address'), required=False,
        constraint=re.compile(
            "^[/$!%=+A-Za-z0-9_.-]+@([A-Za-z0-9_\-]+\.)+[A-Za-z]{2,6}$").match)


class IHomePageAddress(IPersonEntity):
    """A home page address."""

    url = zope.schema.URI(title=_(u'URL'), required=False)


class IPhoneNumber(IPersonEntity):
    """A phone number."""

    number = zope.schema.TextLine(title=_(u'number'), required=False)


class IPersonDefaults(zope.interface.Interface):
    """Default addresses, phone numbers etc."""

    default_postal_address = zope.schema.Choice(
        title=_(u'main postal address'),
        source=ContextByInterfaceSource(IPostalAddress))
    default_phone_number = zope.schema.Choice(
        title=_(u'main phone number'),
        source=ContextByInterfaceSource(IPhoneNumber))
    default_email_address = zope.schema.Choice(
        title=_(u'main e-mail address'),
        source=ContextByInterfaceSource(IEMailAddress))
    default_home_page_address = zope.schema.Choice(
        title=_(u'main home page address'),
        source=ContextByInterfaceSource(IHomePageAddress))


class IEntities(zope.interface.Interface):
    """Entities in the address book."""

    def getEntities(sorted=True):
        """Get an iterable of all known entities.

        When `sorted` is True, order them as defined in IEntityOrder.

        """

    def getMainEntities(sorted=True):
        """Get an iterable of the most important entities.

        When `sorted` is True, order them as defined in IEntityOrder.

        """


class IEntityRead(zope.interface.Interface):
    """Entity in the address book. (Read only part of the interface.)"""

    title = zope.schema.TextLine(
        title=u"title of the entity", required=False)
    interface = zope.schema.InterfaceField(title=u"interface")
    class_name = zope.schema.DottedName(
        title=u"dotted name of the class", required=False)

    name = zope.interface.Attribute(
        "Uniqe name of the entity which only contains letters.")
    tagged_values = zope.interface.Attribute(
        "Dict of tagged values of the entity.")
    order_storage_namespace = zope.interface.Attribute(
        "Get the name space used in the order storage.")

    def getRawField(field_name):
        """Get a field by its name. (Without any conversion.)"""

    def getField(field_name):
        """Get a zope.schema field by its name.

        When the field is a user defined one it gets converted to a zope.schema
        field.
        """

    def getFieldOrder():
        """Get the ordered names of the fields.

        Caution: This method only returns the names for the fields which are
        known in the field order!
        """

    def getRawFields(sorted=True):
        """Get (name, field) tuples of the schema fields on the entity.

        When `sorted` is true the fields are sorted using the field order.

        Returns static (zope.schema) and user defined (IField) fields.

        """

    def getFields(sorted=True):
        """Get (name, field) tuples of the schema fields on the entity.

        When `sorted` is true the fields are sorted using the field order.

        Converts user defined fields (see IField) into zope.schema fields.

        """

    def getFieldValues(sorted=True):
        """Get list of the schema fields on the entity.

        When `sorted` is true the fields are sorted using the field order.

        Converts user defined fields (see IField) into zope.schema fields.

        """

    def getClass():
        """Get the class object for `self.class_name`."""


class IEntityWrite(zope.interface.Interface):
    """Entity in the address book. (Write part of the interface.)"""

    def addField(field):
        """Add a user defined field to the entity.

        Returns the name of the created field.

        """

    def removeField(field):
        """Remove a user defined field from the entity."""

    def setFieldOrder(field_names):
        """Update the order of the fields like in `field_names`.

        field_names ... list of the names of the fields as returned by
                        `getFields`.

        Field names which do not belong to the entity are omitted.
        """


class IEntity(IEntityRead, IEntityWrite):
    """Entity in the address book."""


class IEditableEntity(IEntity):
    """Special entity which is editable.

    This means that new fields can be added and the fields can be sorted.

    """


class IField(zope.interface.Interface):
    """User defined field."""

    __name__ = zope.interface.Attribute('internal field name')
    interface = zope.interface.Attribute('interface the field belongs to')

    type = zope.schema.Choice(
        title=_(u'type'),
        source=icemac.addressbook.sources.FieldTypeSource())
    title = zope.schema.TextLine(title=_(u'title'))
    values = zope.schema.List(
        title=_(u'choice field values'), unique=True, required=False,
        value_type=zope.schema.TextLine(title=_(u'value')))
    notes = zope.schema.Text(title=_(u'notes'), required=False)

    @zope.interface.invariant
    def choice_type_needs_values(field):
        if field.type == u'Choice' and not field.values:
            raise zope.interface.Invalid(
                _(u'type "choice" requires at least one field value.'))


class IMayHaveUserFields(zope.interface.Interface):
    """Marker interface: Object may have user defined fields."""


class IUserFieldStorage(zope.interface.Interface):
    """Storage for user defined field values."""


class IOrderStorageRead(zope.interface.Interface):
    """Storage of orders of objects. (read only methods)"""

    def namespaces():
        """Get an iterable of the known namespaces."""

    def get(obj, namespace):
        """Get the index of the object in the list.

        Raises a KeyError when the object is not in the list.
        """

    def isFirst(obj, namespace):
        """Tell whether `obj` is the first object in the list.

        Raises a KeyError when the object is not in the list.
        """

    def isLast(obj, namespace):
        """Tell whether `obj` is the last object in the list.

        Raises a KeyError when the object is not in the list.
        """

    def byNamespace(namespace):
        """List of the objects registered for the namespace."""


class IOrderStorageWrite(zope.interface.Interface):
    """Storage of orders of objects. (write methods)"""

    def add(obj, namespace):
        """Add an object to the order for a namespace.

        When the namespace does not exist, it gets created.
        New objects are added at the end of the list.
        When the object is already in the list it is not added again and
        does not change its position.
        """

    def remove(obj, namespace):
        """Remove the object from the order of a namespace.

        Raises a ValueError when the object is not in the list.
        Raises a KeyError when the namespace does not exist.
        """

    def truncate(namespace):
        """Remove all objects from the order of a namespace.

        Does nothing when the namespace is not known.
        """

    def up(obj, namespace, delta=1):
        """Move the object one position up in the list.

        `delta` describes the number of positions to move.

        When it would be moved beyond the beginning of the list a
        ValueError is raised.
        """

    def down(obj, namespace, delta=1):
        """Move the object one position down in the list.

        `delta` describes the number of positions to move.

        When it would be moved beyond the end of the list a ValueError is
        raised.
        """


class IOrderStorage(IOrderStorageRead, IOrderStorageWrite):
    """Storage of orders of objects."""


class IEntityOrder(zope.interface.Interface):
    """Order of entities."""

    def get(entity):
        """Get the index of the entity in the entity order.

        Raises a KeyError when the entity is not known.
        """

    def isFirst(entity):
        """Tell whether `entity` comes first in the entity order.

        Raises a KeyError when the entity is not known.
        """

    def isLast(entity):
        """Tell whether `entity` comes last in the entity order.

        Raises a KeyError when the entity is not known.
        """

    def __iter__():
        """Iterate over the entities sorted by order."""

    def up(entity, delta=1):
        """Move the entity one position up in the entity order.

        `delta` describes the number of positions to move.

        When the entity would be moved beyond the beginning of the entity
        order a ValueError is raised.
        """

    def down(entity, delta=1):
        """Move the entity one position down in the entity order.

        `delta` describes the number of positions to move.

        When it would be moved beyond the end of the entity order a
        ValueError is raised.
        """



import zope.interface
import icemac.addressbook.export.interfaces


class BaseExporter(object):
    """Abstract base for exporters which defines some convenience methods."""

    zope.interface.implements(icemac.addressbook.export.interfaces.IExporter)

    # to be set in subclass
    file_extension = NotImplemented
    mime_type = NotImplemented

    def __init__(self, persons, request=None):
        self.persons = persons
        self.request = request

    def export(self):
        raise NotImplementedError()



import icemac.addressbook.export.base
import icemac.addressbook.export.interfaces
import pytest
import zope.interface.verify


def test_base__BaseExport__1():
    """`BaseExport` conforms to `IExporter`."""
    assert zope.interface.verify.verifyObject(
        icemac.addressbook.export.interfaces.IExporter,
        icemac.addressbook.export.base.BaseExporter([], None))


def test_base__BaseExport__export__1():
    """`export` is not implemented."""
    # Test is only here to get better test coverage.
    exporter = icemac.addressbook.export.base.BaseExporter(None)
    with pytest.raises(NotImplementedError):
        exporter.export()



import pytz
import zope.component
import zope.preference.interfaces


def get_preference_group(id):
    return zope.component.getUtility(
        zope.preference.interfaces.IPreferenceGroup, name=id)


def get_time_zone_name():
    """User selected time zone name."""
    return get_preference_group('ab.timeZone').time_zone


def get_time_zone():
    """User selected time zone object."""
    return pytz.timezone(get_time_zone_name())



# -*- coding: utf-8 -*-
import pytest
import urllib2
from icemac.addressbook.testing import Browser


@pytest.mark.parametrize('username', ('visitor', 'editor'))
def test_update__security__1(search_data, browser, username):
    """Some roles are not able to see the update search result handler."""
    browser.login(username)
    browser.keyword_search('church')
    assert ['XLS export main (Exports person data and main addresses resp. '
            'phone numbers.)',
            'XLS export complete (Exports person data and all addresses '
            'resp. phone numbers.)',
            'E-Mail (Creates a link to send e-mails.)',
            'Names (Comma separated list of person names.)',
            'Checklist (List of person names with check-boxes.)',
            "iCalendar export birthday (Export person's birthdays as .ics "
            "file.)"] == browser.getControl(
                'Apply on selected persons').displayOptions


@pytest.mark.parametrize(
    'url', (Browser.SEARCH_MULTI_UPDATE_URL,
            Browser.SEARCH_MULTI_UPDATE_CHOOSE_FIELD_URL,
            Browser.SEARCH_MULTI_UPDATE_ENTER_VALUE_URL,
            Browser.SEARCH_MULTI_UPDATE_CHECK_RESULT_URL,
            Browser.SEARCH_MULTI_UPDATE_COMPLETE_URL))
@pytest.mark.parametrize('username', ('visitor', 'editor'))
def test_update__security__2(search_data, browser, url, username):
    """Some roles are not able to access update search result handler URLs.

    This is even right when he knows the URLs.

    """
    browser.login(username)
    with pytest.raises(urllib2.HTTPError) as err:
        browser.open(url)
    assert 'HTTP Error 403: Forbidden' == str(err.value)



from icemac.addressbook.address import phone_number_entity
from icemac.addressbook.address import postal_address_entity
from icemac.addressbook.addressbook import address_book_entity
from icemac.addressbook.entities import Entities, PersistentEntities, Field
from icemac.addressbook.entities import Entity, get_bound_schema_field
from icemac.addressbook.entities import EntityOrder
from icemac.addressbook.interfaces import IAddressBook, IHomePageAddress
from icemac.addressbook.interfaces import IEntities, IEntityOrder, IEntity
from icemac.addressbook.interfaces import IField, IOrderStorage
from icemac.addressbook.interfaces import IPhoneNumber, IPerson, IKeyword
from icemac.addressbook.interfaces import IUserFieldStorage
from icemac.addressbook.tests.conftest import IDog, IKwack, Kwack
import persistent
import pytest
import zope.component
import zope.component.hooks
import zope.interface
import zope.schema
import zope.schema.interfaces


# Constants

DEFAULT_ORDER = [u'person',
                 u'postal address',
                 u'phone number',
                 u'e-mail address',
                 u'home page address']


# Fixtures

@pytest.fixture(scope='function')
def entityOrder(address_book):
    """Get the entity order utility."""
    return zope.component.getUtility(IEntityOrder)


@pytest.fixture(scope='module')
def unknownEntity():
    """Get an entity which is not known in the address book."""
    return Entity(u'', IEntities, 'icemac.addressbook.entities.Entities')


@pytest.fixture(scope='function')
def minimalEntity(address_book):
    """Get an entity which is not registered in the address book."""
    return IEntity(IEntities)


@pytest.fixture(scope='function')
def field():
    """Get a user defined field."""
    field = Field()
    field.type = 'TextLine'
    return field


@pytest.fixture(scope='function')
def schemaized_field(field):
    """Get a user defined field adapted to a `zope.schema` field."""
    return zope.schema.interfaces.IField(field)


@pytest.fixture(scope='function')
def entity():
    """Get a custom entity."""
    return Entity(
        u'Dummy', IDummy, 'icemac.addressbook.tests.test_entities.Dummy')


@pytest.fixture(scope='function')
def entity_with_field(address_book, entity, field):
    """Get a custom entity with a user defined field."""
    entity.addField(field)
    return entity


@pytest.fixture(scope='function')
def entities(address_book):
    """Get the entities utility."""
    return zope.component.getUtility(IEntities)


# Helper classes


class IDummy(zope.interface.Interface):
    """Interface for test entity."""

    dummy = zope.schema.Text(title=u'dummy')
    dummy2 = zope.schema.Text(title=u'dummy2')


class Dummy(object):
    """Test entity."""

    zope.interface.implements(IDummy)


# Helper functions

def getMainEntities_titles(sorted):
    """Return the entities titles when calling `getMainEntities()`."""
    entities = zope.component.getUtility(IEntities)
    return [x.title for x in entities.getMainEntities(sorted=sorted)]


# Tests


def test_entities__Entities__1():
    """`Entities` fulfills the `IEntities` interface."""
    zope.interface.verify.verifyObject(IEntities, Entities())


def test_entities__PersistentEntities__1():
    """`PersistentEntities` fulfills the `IEntities` interface."""
    zope.interface.verify.verifyObject(IEntities, PersistentEntities())


def test_entities__EntityOrder__1():
    """`EntityOrder` fulfills the `IEntityOrder` interface."""
    zope.interface.verify.verifyObject(IEntityOrder, EntityOrder())


def test_entities__Field__1():
    """`Field` fulfills the `IField` interface."""
    zope.interface.verify.verifyObject(IField, Field())


def test_entities__Entity__1():
    """`Entity` fulfills the `IEntity` interface."""
    entity = Entity(None, IEntity, 'Entity')
    zope.interface.verify.verifyObject(IEntity, entity)


def test_entities__getEntities__1(stubSortOrder, stubEntities):
    """`(Persistent)Entities.getEntities()` returns entities sorted.

    The sort order is defined in the sort order storage.
    """
    e = stubEntities
    assert [e.cat, e.kwack, e.duck] == e.entities.getEntities()


def test_entities__getEntities__2(stubEntities):
    """`(Persistent)Entities.getEntities()` might return entities unsorted."""
    e = stubEntities
    assert (sorted([e.kwack, e.duck, e.cat]) ==
            sorted(e.entities.getEntities(sorted=False)))


def test_entities__getEntities__3(stubSortOrder, stubEntities):
    """`(Persistent)Entities.getEntities()` respects changes in sort order."""
    e = stubEntities
    stubSortOrder.up(e.duck)
    assert [e.cat, e.duck, e.kwack] == e.entities.getEntities()


def test_entities__PersistentEntities__getMainEntities__1(address_book):
    """By default the result of `getMainEntities()` is sort independent."""
    assert DEFAULT_ORDER == getMainEntities_titles(True)
    assert DEFAULT_ORDER == getMainEntities_titles(False)


def test_entities__PersistentEntities__getMainEntities__2(entityOrder):
    """`getMainEntities()` respects a changed sort order."""
    entityOrder.up(IEntity(IPhoneNumber))
    assert ([u'person',
             u'phone number',
             u'postal address',
             u'e-mail address',
             u'home page address'] == getMainEntities_titles(True))
    # The unordered variant still returns the default order:
    assert DEFAULT_ORDER == getMainEntities_titles(False)


def test_entities__EntityOrder__get__1(entityOrder):
    """`get()` returns the position of the entity."""
    assert 1 == entityOrder.get(IEntity(IPerson))
    assert 8 == entityOrder.get(IEntity(IKeyword))


def test_entities__EntityOrder__get__2(entityOrder, unknownEntity):
    """`get()` raises a KeyError if entity is not known to the sort order."""
    with pytest.raises(KeyError):
        entityOrder.get(unknownEntity)


def test_entities__EntityOrder__get__3(entityOrder, minimalEntity):
    """`get()` raises a KeyError if the entity has no name."""
    with pytest.raises(KeyError):
        entityOrder.get(minimalEntity)


def test_entities__EntityOrder__get__4(entityOrder, AddressBookFactory):
    """It accesses the address book in `zope.component.hooks.site`."""
    other_address_book = AddressBookFactory('other_address_book')
    # Changes in an address book ...
    person = IEntity(IPerson)
    assert 1 == entityOrder.get(person)
    entityOrder.down(person)
    assert 2 == entityOrder.get(person)
    # ... are not reflected in another address book:
    with zope.component.hooks.site(other_address_book):
        assert 1 == entityOrder.get(person)


def test_entities__EntityOrder__get__5(entityOrder):
    """`get()` raises a ComponentLookupError if no site is set."""
    person = IEntity(IPerson)
    with zope.component.hooks.site(None):
        with pytest.raises(zope.component.ComponentLookupError):
            entityOrder.get(person)


def test_entities__EntityOrder__isFirst__1(entityOrder):
    """`isFirst()` returns `True` for the first entity."""
    assert entityOrder.isFirst(IEntity(IAddressBook))


def test_entities__EntityOrder__isFirst__2(entityOrder):
    """`isFirst()` returns `False` for an entity which is not the first."""
    assert not entityOrder.isFirst(IEntity(IPhoneNumber))


def test_entities__EntityOrder__isFirst__3(entityOrder, unknownEntity):
    """`isFirst()` raises KeyError if entity is not known to the sort order."""
    with pytest.raises(KeyError):
        entityOrder.isFirst(unknownEntity)


def test_entities__EntityOrder__isFirst__4(entityOrder, minimalEntity):
    """`isFirst()` raises KeyError if the entity has no name."""
    with pytest.raises(KeyError):
        entityOrder.isFirst(minimalEntity)


def test_entities__EntityOrder__isLast__1(entityOrder):
    """`isLast()` returns `True` for the first entity."""
    assert entityOrder.isLast(IEntity(IKeyword))


def test_entities__EntityOrder__isLast__2(entityOrder):
    """`isLast()` returns `False` for an entity which is not the first."""
    assert not entityOrder.isLast(IEntity(IPhoneNumber))


def test_entities__EntityOrder__isLast__3(entityOrder, unknownEntity):
    """`isLast()` raises KeyError if entity is not known to the sort order."""
    with pytest.raises(KeyError):
        entityOrder.isLast(unknownEntity)


def test_entities__EntityOrder__isLast__4(entityOrder, minimalEntity):
    """`isLast()` raises KeyError if the entity has no name."""
    with pytest.raises(KeyError):
        entityOrder.isLast(minimalEntity)


def test_entities__EntityOrder____iter____1(entityOrder):
    """It is iterable, returning entity names."""
    assert ([
        'IcemacAddressbookAddressbookAddressbook',
        'IcemacAddressbookPersonPerson',
        'IcemacAddressbookPersonPersondefaults',
        'IcemacAddressbookAddressPostaladdress',
        'IcemacAddressbookAddressPhonenumber',
        'IcemacAddressbookAddressEmailaddress',
        'IcemacAddressbookAddressHomepageaddress',
        'IcemacAddressbookFileFileFile',
        'IcemacAddressbookKeywordKeyword',
    ] == list(iter(entityOrder)))


def test_entities__EntityOrder__up__1(entityOrder):
    """`up()` moved one position up."""
    person = IEntity(IPerson)
    assert 1 == entityOrder.get(person)
    entityOrder.up(person)
    assert 0 == entityOrder.get(person)


def test_entities__EntityOrder__up__2(entityOrder):
    """`up(n)` moves n positions up."""
    hp = IEntity(IHomePageAddress)
    assert 6 == entityOrder.get(hp)
    entityOrder.up(hp, 3)
    assert 3 == entityOrder.get(hp)


def test_entities__EntityOrder__up__3(entityOrder):
    """If the delta is too big when calling `up()` a ValueError is raised."""
    person = IEntity(IPerson)
    assert 1 == entityOrder.get(person)
    with pytest.raises(ValueError):
        entityOrder.up(person, 2)


def test_entities__EntityOrder__down__1(entityOrder):
    """`down()` moved one position down."""
    person = IEntity(IPerson)
    assert 1 == entityOrder.get(person)
    entityOrder.down(person)
    assert 2 == entityOrder.get(person)


def test_entities__EntityOrder__down__2(entityOrder):
    """`down(n)` moves n positions down."""
    ab = IEntity(IAddressBook)
    assert 0 == entityOrder.get(ab)
    entityOrder.down(ab, 3)
    assert 3 == entityOrder.get(ab)


def test_entities__EntityOrder__down__3(entityOrder):
    """If the delta is too big when calling `down()` a ValueError is raised."""
    person = IEntity(IPerson)
    assert 1 == entityOrder.get(person)
    with pytest.raises(ValueError):
        entityOrder.down(person, 8)


def test_entities__entity_by_name__1(stubEntities, entityAdapters):
    """`entity_by_name` raises a `ValueError` if string name is unknown."""
    with pytest.raises(ValueError):
        IEntity('asdf')


def test_entities__entity_by_name__2(stubEntities, entityAdapters):
    """`entity_by_name` raises a `ValueError` if unicode name is unknown."""
    with pytest.raises(ValueError):
        IEntity(u'asdf')


def test_entities__entity_by_name__3(stubEntities, entityAdapters):
    """`entity_by_name` raises a `ValueError` if class name is used."""
    with pytest.raises(ValueError):
        IEntity('icemac.addressbook.tests.stubs.Duck')


def test_entities__entity_by_name__4(stubEntities, entityAdapters):
    """`entity_by_name` raises a `ValueError` if class name is used."""
    with pytest.raises(ValueError):
        IEntity('icemac.addressbook.tests.conftest.Duck')


def test_entities__entity_by_name__5(stubEntities, entityAdapters):
    """`entity_by_name` adapts from string entity name."""
    assert stubEntities.duck == IEntity('IcemacAddressbookTestsConftestDuck')


def test_entities__entity_by_name__6(stubEntities, entityAdapters):
    """`entity_by_name` adapts from unicode entity name."""
    assert stubEntities.duck == IEntity(u'IcemacAddressbookTestsConftestDuck')


def test_entities__entity_by_interface__1(stubEntities, entityAdapters):
    """`entity_by_interface` returns a minimal entity for unknown interface."""
    entity = IEntity(IDog)
    assert None is entity.title
    assert IDog is entity.interface
    assert None is entity.class_name


def test_entities__entity_by_interface__2(stubEntities, entityAdapters):
    """`entity_by_interface` returns the entity for a known interface."""
    assert stubEntities.kwack == IEntity(IKwack)


def test_entities__entity_by_obj__1(stubEntities, entityAdapters):
    """`entity_by_obj` raises a ValueError for an unknown instance."""
    obj = persistent.Persistent()
    with pytest.raises(ValueError):
        IEntity(obj)


def test_entities__entity_by_obj__2(stubEntities, entityAdapters):
    """`entity_by_obj` returns the instance for a known instance."""
    assert stubEntities.kwack == IEntity(Kwack())


def test_entities__Entity__addField__1(entities, entity, field):
    """`addField` adds a field to an entity."""
    entity.addField(field)
    assert ['Field-1'] == list(entities.keys())
    assert field is entities[u'Field-1']


def test_entities__Entity__addField__2(entities, entity, field):
    """`addField` stores interface of the entity on the field."""
    entity.addField(field)
    assert IDummy == entities[u'Field-1'].interface


def test_entities__Entity__addField__3(entities, entity, field):
    """`addField` registers the field as an adapter."""
    entity.addField(field)
    field_adapter = zope.component.getMultiAdapter(
        (entity, Dummy()), IField, name=u'Field-1')
    assert field is field_adapter


def test_entities__Entity__removeField__1(entities, entity_with_field, field):
    """`removeField`  removes a field from an entity."""
    assert field in entities.values()
    entity_with_field.removeField(field)
    assert field not in entities.values()


def test_entities__Entity__removeField__2(entity_with_field, field):
    """`removeField` removes the interface of the entity from the field."""
    entity_with_field.removeField(field)
    assert field.interface is None


def test_entities__Entity__removeField__3(entity_with_field, field):
    """`removeField` removes the adapter registration of the field."""
    entity_with_field.removeField(field)
    assert None is zope.component.queryMultiAdapter(
        (entity, Dummy()), IField, name=u'Field')


def test_entities__Entity__setFieldOrder__1(entity_with_field, field):
    """`setFieldOrder` changes the initial (empty) order."""
    assert [] == entity_with_field.getFieldOrder()
    entity_with_field.setFieldOrder(['dummy2', field.__name__, 'dummy'])
    assert (['dummy2', field.__name__, 'dummy'] ==
            entity_with_field.getFieldOrder())


def test_entities__Entity__setFieldOrder__2(entity_with_field):
    """`setFieldOrder` ignores unknown field names and does not store them."""
    entity = entity_with_field
    entity.setFieldOrder(['dummy2', 'I-do-not-exist', 'dummy'])
    assert ['dummy2', 'dummy'] == entity.getFieldOrder()
    # Unknown field names are not written into storage:
    order_storage = zope.component.getUtility(IOrderStorage)
    assert (['dummy2', 'dummy'] ==
            order_storage.byNamespace(entity.order_storage_namespace))


def test_entities__Entity__getFieldOrder__1(entity):
    """`getFieldOrder` returns the initially emtpy field order."""
    assert [] == entity.getFieldOrder()


def test_entities__Entity__getFieldOrder__2(entity_with_field, field):
    """`getFieldOrder` only contains the values set by `setFieldOrder`."""
    entity = entity_with_field
    entity.setFieldOrder([field.__name__, 'dummy'])
    assert [field.__name__, 'dummy'] == entity.getFieldOrder()


def test_entities__Entity__getFieldOrder__3(address_book):
    """`getFieldOrder` returns `[]` if the namespace cannot be computed."""
    # The namespace in the order utility depends on the name of the
    # entity which itself depends on the class_name stored on the
    # entity. But this class name is optional, so the name might not be
    # computable:
    entity = Entity(u'Dummy', IDummy, None)
    assert [] == entity.getFieldOrder()


def test_entities__Entity__name__1(entity):
    """`name` is the name of the entity containing the module path."""
    assert 'IcemacAddressbookTestsTestEntitiesDummy' == entity.name


def test_entities__Entity__name__2(address_book):
    """`name` raises a ValueError if no class name is set."""
    entity = Entity(None, IDummy, None)
    with pytest.raises(ValueError):
        entity.name


def test_entities__Entity__getRawFields__1(entity_with_field, field):
    """`getRawFields` sorts fields missing in sort order to the end."""
    entity = entity_with_field
    entity.setFieldOrder([field.__name__, 'dummy'])
    assert ([field.__name__, 'dummy', 'dummy2'] ==
            [x[0] for x in entity.getRawFields()])


def test_entities__Entity__getRawFields__2(entity_with_field, field):
    """`getRawFields` returns the fields sorted by the field order."""
    entity = entity_with_field
    entity.setFieldOrder(['dummy2', field.__name__, 'dummy'])
    assert ([('dummy2', IDummy['dummy2']),
             (field.__name__, field),
             ('dummy', IDummy['dummy'])] == list(entity.getRawFields()))


def test_entities__Entity__getRawFields__3(entity_with_field, field):
    """`getRawFields` is able to return the fields unsorted (default order).

    When `sorted` is `False` the `zope.schema` fields are returned first (the
    order is defined by the order in the interface) and than the user defined
    fields (order is undefined here.)
    """
    entity = entity_with_field
    entity.setFieldOrder(['dummy2', 'Field', 'dummy'])
    assert ([('dummy', IDummy['dummy']),
             ('dummy2', IDummy['dummy2']),
             (field.__name__, field)] ==
            list(entity.getRawFields(sorted=False)))


def test_entities__Entity__getFields__1(entity_with_field, schemaized_field):
    """`getFields` returns all fields as `zope.schema` in order."""
    entity = entity_with_field
    entity.setFieldOrder(['dummy2', schemaized_field.__name__, 'dummy'])
    assert ([('dummy2', IDummy['dummy2']),
             (schemaized_field.__name__, schemaized_field),
             ('dummy', IDummy['dummy'])] == list(entity.getFields()))


def test_entities__Entity__getFields__2(entity_with_field, schemaized_field):
    """`geFields` is able to return the fields unsorted (default order).

    When `sorted` is `False` the `zope.schema` fields are returned first (the
    order is defined by the order in the interface) and then the user defined
    fields (order is undefined here.)
    """
    entity = entity_with_field
    entity.setFieldOrder(['dummy2', schemaized_field.__name__, 'dummy'])
    assert ([('dummy', IDummy['dummy']),
             ('dummy2', IDummy['dummy2']),
             (schemaized_field.__name__, schemaized_field)] ==
            list(entity.getFields(sorted=False)))


def test_entities__Entity__getFieldValues__1(
        entity_with_field, schemaized_field):
    """`getFieldValues` returns all fields as `zope.schema` in order."""
    entity = entity_with_field
    entity.setFieldOrder(['dummy2', schemaized_field.__name__, 'dummy'])
    assert ([IDummy['dummy2'], schemaized_field, IDummy['dummy']] ==
            entity.getFieldValues())


def test_entities__Entity__getFieldValues__2(
        entity_with_field, schemaized_field):
    """`getFieldValues` is able to return the fields unsorted (default order).

    When `sorted` is `False` the `zope.schema` fields are returned first (the
    order is defined by the order in the interface) and then the user defined
    fields (order is undefined here.)
    """
    entity = entity_with_field
    entity.setFieldOrder(['dummy2', schemaized_field.__name__, 'dummy'])
    assert ([IDummy['dummy'], IDummy['dummy2'], schemaized_field] ==
            entity.getFieldValues(sorted=False))


def test_entities__Entity__getField__1(entity):
    """`getField` raises a `KeyError` for an unknown field name."""
    with pytest.raises(KeyError):
        entity.getField('asdf')


def test_entities__Entity__getField__2(entity):
    """`getField` can return a `zope.schema` field."""
    assert IDummy['dummy2'] == entity.getField('dummy2')


def test_entities__Entity__getField__3(entity_with_field, schemaized_field):
    """`getField` can return a user defined field as `zope.schema` field."""
    assert (schemaized_field ==
            entity_with_field.getField(schemaized_field.__name__))


def test_entities__Entity__getRawField__1(entity):
    """`getRawField` raises a `KeyError` for an unknown field name."""
    with pytest.raises(KeyError):
        entity.getRawField('asdf')


def test_entities__Entity__getRawField__2(entity):
    """`getRawField` can return a `zope.schema` field."""
    assert IDummy['dummy2'] == entity.getRawField('dummy2')


def test_entities__Entity__getRawField__3(entity_with_field, field):
    """`getRawField` can return a user defined field."""
    assert field == entity_with_field.getRawField(field.__name__)


def test_entities__Entity__getClass__1(entity):
    """`getClass` returns the class object associated with the entity."""
    assert Dummy == entity.getClass()


def test_entities__Entity__getClass__2():
    """`getClass` raises a `ValueError` if no `class_name` is set."""
    e = Entity(None, IDummy, None)
    with pytest.raises(ValueError):
        e.getClass()


def test_entities__Entity__tagged_values__1():
    """`tagged_values` is a dict of the kwargs used during entity __init__."""
    e = Entity(u'Dummy', IDummy, 'Dummy', a=1, b='asdf')
    assert dict(a=1, b='asdf') == e.tagged_values


def test_entities__Entity__tagged_values__2():
    """`tagged_values` is actually a copy of the kwargs dict.

    Tagged values not modifiable by modifying the returned dict.
    """
    e = Entity(u'Dummy', IDummy, 'Dummy', a=1, b='asdf')
    e.tagged_values['a'] = 2
    assert dict(a=1, b='asdf') == e.tagged_values


def test_entities__get_bound_schema_field__1(address_book):
    """It returns the schema field if the obj provides the entity interface."""
    field = get_bound_schema_field(address_book, address_book_entity,
                                   address_book_entity.getRawField('title'))
    assert address_book == field.context
    assert 'title' == field.__name__
    assert isinstance(field, zope.schema.TextLine)


def test_entities__get_bound_schema_field__2(address_book, FullPersonFactory):
    """It looks up the default obj on obj if iface is not provided by obj."""
    person = FullPersonFactory(address_book, u'Koch')
    field = get_bound_schema_field(
        person, postal_address_entity,
        postal_address_entity.getRawField('country'))
    assert person.default_postal_address == field.context
    assert 'country' == field.__name__
    assert isinstance(field, zope.schema.Choice)


def test_entities__get_bound_schema_field__3(address_book, FullPersonFactory):
    """It binds field to obj if `default_attrib_fallback` is `False`.

    Additional condition: The interface is not provided by obj.
    """
    person = FullPersonFactory(address_book, u'Koch')
    field = get_bound_schema_field(
        person, postal_address_entity,
        postal_address_entity.getRawField('country'),
        default_attrib_fallback=False)
    assert person == field.context


def test_entities__get_bound_schema_field__4(
        address_book, FullPersonFactory, FieldFactory):
    """It returns a user defined field as a `zope.schema` field."""
    person = FullPersonFactory(address_book, u'Koch')
    FieldFactory(address_book, IPhoneNumber, 'Datetime', u'last call',
                 notes=u'using a phone')
    field = get_bound_schema_field(
        person, phone_number_entity,
        phone_number_entity.getRawField('Field-1'))
    assert IUserFieldStorage(person.default_phone_number) == field.context
    assert 'Field-1' == field.__name__
    assert 'using a phone' == field.description
    assert isinstance(field, zope.schema.Datetime)



from mock import patch
from z3c.flashmessage.interfaces import IMessageSource
import datetime
import pytest
import zope.component
import zope.publisher.testing


@pytest.yield_fixture('function')
def fake_session(empty_zodb):
    """Fake a user session to be static."""
    with patch('zope.session.interfaces.ISession') as ISession:
        ISession.return_value = {'z3c.flashmessage': {}}
        yield


def send_msg(msg):
    """Send a flash message."""
    with zope.publisher.testing.interaction('zope.mgr'):
        zope.component.getUtility(IMessageSource).send(msg)


def test_testing__Browser__message__1(fake_session, browser):
    """It is an empty list if there are no flash messages."""
    browser.login('mgr')
    browser.open('http://localhost')
    assert [] == browser.message


def test_testing__Browser__message__2(fake_session, browser):
    """It is a string if there is exactly one flash message."""
    send_msg('foo')
    browser.login('mgr')
    browser.handleErrors = False
    browser.open('http://localhost')
    assert 'foo' == browser.message


def test_testing__Browser__message__3(fake_session, browser):
    """It is a list of strings if there is more than one flash message."""
    send_msg('foo')
    send_msg('blah')
    browser.login('mgr')
    browser.open('http://localhost')
    assert ['foo', 'blah'] == browser.message


def test_testing__DateTimeClass__format__1(DateTime):
    """It formats a date as zope.testbrowser expects it."""
    assert '2017 1 27 ' == DateTime.format(datetime.date(2017, 1, 27))


def test_testing__SeleniumLogin__1(address_book, browser):
    """It cannot be accessed by an anonymous user."""
    with pytest.raises(zope.testbrowser.browser.HTTPError) as err:
        browser.open(browser.SELENIUM_LOGIN_URL)
    assert 'HTTP Error 401: Unauthorized' == str(err.value)



import icemac.addressbook.generations.utils

# needed to make sure this package still exists and the instances get
# converted to the new class path
import zope.app.authentication

a = zope.app.authentication


def evolve(context):
    """Update the authentication utility."""
    icemac.addressbook.generations.utils.update_address_book_infrastructure(
        context)
